# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
import operator

import torch
import torch.nn as nn

import brevitas
from brevitas.graph.base import InsertModuleCallAfter
from brevitas.graph.base import ModuleInstanceToModuleInstance
from brevitas.graph.base import ModuleToModuleByInstance
from brevitas.graph.utils import get_module

ADD_FNS = [torch.add, operator.add, operator.iadd]

ADD_METHODS = ['add', 'add_']
CAT = brevitas.original_cat

SIGN_PRESERVING_MODULES = (
    nn.Dropout,
    nn.Dropout2d,
    nn.Dropout3d,
    nn.MaxPool1d,
    nn.MaxPool2d,
    nn.MaxPool3d,
    nn.AvgPool1d,
    nn.AvgPool2d,
    nn.AvgPool3d,
    nn.AdaptiveAvgPool1d,
    nn.AdaptiveAvgPool2d,
    nn.AdaptiveAvgPool3d)

PRECISION_PRESERVING_MODULES = (
    nn.Dropout, nn.Dropout2d, nn.Dropout3d, nn.MaxPool1d, nn.MaxPool2d, nn.MaxPool3d)


def inp_placeholder_handler(model, input_quantizer):
    """
    Add Quantization step at the input of the network.
    """
    rewriters = []
    if input_quantizer is None:
        return model
    for node in model.graph.nodes:
        if node.op == 'placeholder':
            act_quant, kwargs_act_quant = input_quantizer
            inp_quant = act_quant(**kwargs_act_quant)
            name = node.name + '_quant'
            model.add_module(name, inp_quant)
            rewriters.append(InsertModuleCallAfter(name, node))
    for rewriter in rewriters:
        model = rewriter.apply(model)
    return model


def are_inputs_unsigned(model, node, is_unsigned_list, quant_act_map, unsigned_act_tuple):
    for inp_node in node.all_input_nodes:
        if inp_node.op == 'call_module':
            inp_module = get_module(model, inp_node.target)
            if isinstance(inp_module, tuple(quant_act_map.keys())) and isinstance(
                    inp_module, unsigned_act_tuple):
                is_unsigned_list.append(True)
            elif isinstance(inp_module, tuple(SIGN_PRESERVING_MODULES)):
                are_inputs_unsigned(
                    model, inp_node, is_unsigned_list, quant_act_map, unsigned_act_tuple)
            elif hasattr(inp_module, 'is_quant_act_signed'):
                is_unsigned_list.append(not inp_module.is_quant_act_signed)
            else:
                is_unsigned_list.append(False)
        elif inp_node.op == 'call_function':
            if inp_node.target in [torch.reshape, torch.flatten, torch.transpose, CAT] + ADD_FNS:
                are_inputs_unsigned(
                    model, inp_node, is_unsigned_list, quant_act_map, unsigned_act_tuple)
            else:
                is_unsigned_list.append(False)
        elif inp_node.op == 'call_method':
            if inp_node.target in ['view', 'reshape', 'flatten', 't', 'permute'] + ADD_METHODS:
                are_inputs_unsigned(
                    model, inp_node, is_unsigned_list, quant_act_map, unsigned_act_tuple)
            else:
                is_unsigned_list.append(False)
    return all(is_unsigned_list)


def _tensor_quant_in_list(act_quant, module_list, same_sign):
    tq = act_quant.fused_activation_quant_proxy.tensor_quant
    for m in module_list:
        if m is None:
            continue
        m_tq = m.fused_activation_quant_proxy.tensor_quant
        if same_sign and m_tq is tq:
            return True
        elif not same_sign and m_tq.scaling_impl is tq.scaling_impl and m_tq.int_scaling_impl is tq.int_scaling_impl:
            return True
    return False


def are_inputs_quantized_and_aligned(model, node, quantized_modules_list, quant_act_map, same_sign):
    """
    Check if the inputs to `node` are quantized and aligned.
    If same_sign=True, aligned means that the inputs should have same sign and scale factor.
    Otherwise, they need to have only the same scale factors.
    If none of the previous conditions are met (e.g., FP input, or not aligned scales), the function
    returns False.
    """
    for inp_node in node.all_input_nodes:
        if inp_node.op == 'call_module':
            inp_module = get_module(model, inp_node.target)
            if isinstance(inp_module, tuple(quant_act_map.keys())):
                quantized_modules_list.append(None)
            elif isinstance(inp_module, tuple(PRECISION_PRESERVING_MODULES)) and (
                    not same_sign or
                (same_sign and isinstance(inp_module, tuple(SIGN_PRESERVING_MODULES)))):
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, same_sign)
            elif hasattr(inp_module, 'act_quant'):
                aq = inp_module.act_quant
                if _tensor_quant_in_list(aq, quantized_modules_list, same_sign):
                    continue
                quantized_modules_list.append(aq)
            else:
                quantized_modules_list.append(None)
        elif inp_node.op == 'call_function':
            if inp_node.target in [torch.reshape, torch.flatten, torch.transpose]:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, same_sign)
            elif inp_node.target is CAT:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, True)
            elif inp_node.target in ADD_FNS:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, False)
            else:
                quantized_modules_list.append(None)
        elif inp_node.op == 'call_method':
            if inp_node.target in ['view', 'reshape', 'flatten', 't', 'permute']:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, same_sign)
            elif inp_node.target in ADD_METHODS:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, False)
            else:
                quantized_modules_list.append(None)
    if None in quantized_modules_list:
        return False
    elif len(quantized_modules_list) > 1:
        return False
    else:
        return True


def output_quant_handler(
        model,
        node,
        rewriters,
        is_sign_preserving,
        quant_identity_map,
        quant_act_map=None,
        unsigned_act_tuple=None):
    """
    Starting from `node`, check if any of the users requires requantization (i.e., it does not have
    an act_quant attribute). In that case, the functions adds a requantization step to which all the
    branches are connected. If another branch has its own requantization step, there will be two
    consecutive for that branch.
    """
    if is_sign_preserving and (quant_act_map is None or unsigned_act_tuple is None):
        raise RuntimeError("Missing information for output_quant_handler")
    quant_module = None
    quant_module_name = None
    for user in node.users:
        output_quant = True
        if user.op == 'call_module':
            user_module = get_module(model, user.target)
            if hasattr(user_module, 'act_quant'):
                output_quant = False
        if output_quant:
            if quant_module_name is None and quant_module is None:
                if is_sign_preserving and are_inputs_unsigned(
                        model, node, [], quant_act_map, unsigned_act_tuple):
                    quant_module_class, quant_module_kwargs = quant_identity_map['unsigned']
                else:
                    quant_module_class, quant_module_kwargs = quant_identity_map['signed']
                quant_module = quant_module_class(**quant_module_kwargs)
                quant_module_name = node.name + '_output_quant'
                model.add_module(quant_module_name, quant_module)
                rewriters.append(InsertModuleCallAfter(quant_module_name, node))


def recursive_input_handler(
        model,
        node,
        shared_quant_identity_name,
        shared_quant_identity,
        rewriters,
        quant_identity_map,
        align_input_quant_fn,
        align_sign):
    """
    For a given CAT or ADD node, iterate through its inputs to make sure they are correctly aligned.
    """
    for inp_node in node.all_input_nodes:
        if inp_node.op == 'call_module':
            module = get_module(model, inp_node.target)
            # Precision preserving modules can be safely traversed
            # In case align_sign is True, the modules should also be sign preserving
            if isinstance(module, tuple(PRECISION_PRESERVING_MODULES)) and (
                    not align_sign or
                (align_sign and isinstance(module, tuple(SIGN_PRESERVING_MODULES)))):
                recursive_input_handler(
                    model,
                    inp_node,
                    shared_quant_identity_name,
                    shared_quant_identity,
                    rewriters,
                    quant_identity_map,
                    align_input_quant_fn,
                    align_sign)
            else:
                # Based on the current module, generate an align_output object
                align_output = align_input_quant_fn(
                    module,
                    shared_quant_identity,
                    shared_quant_identity_name,
                    quant_identity_map,
                    align_sign)
                # If it is a tuple, it is a combination of QuantAct and its configuration that
                # will replace the current inp_node module
                if isinstance(align_output, tuple):
                    quant_module_class, quant_module_kwargs = align_output
                    rewriter = ModuleToModuleByInstance(
                        module, quant_module_class, **quant_module_kwargs)
                    rewriters.append(rewriter)
                # If it is a nn.Module, it is an already instatiated QuantAct that will replace
                # the current inp_node module
                elif isinstance(align_output, torch.nn.Module):
                    rewriter = ModuleInstanceToModuleInstance(module, align_output)
                    rewriters.append(rewriter)
                # If it is a string, we simply add a requantization activation after the current
                # inp_node module
                elif isinstance(align_output, str):
                    rewriters.append(InsertModuleCallAfter(shared_quant_identity_name, inp_node))
                else:
                    assert align_output is None, f"align_output {str(align_output)} not supported."
        elif inp_node.op == 'call_function' and inp_node.target in [
                torch.flatten, torch.reshape, torch.transpose, operator.getitem,
                operator.__getitem__]:
            recursive_input_handler(
                model,
                inp_node,
                shared_quant_identity_name,
                shared_quant_identity,
                rewriters,
                quant_identity_map,
                align_input_quant_fn,
                align_sign)
        elif inp_node.op == 'call_function' and inp_node.target is CAT:
            recursive_input_handler(
                model,
                inp_node,
                shared_quant_identity_name,
                shared_quant_identity,
                rewriters,
                quant_identity_map,
                align_input_quant_fn,
                align_sign=True)
        elif inp_node.op == 'call_method' and inp_node.target in [
                'view', 'reshape', 'flatten', 'transpose']:
            recursive_input_handler(
                model,
                inp_node,
                shared_quant_identity_name,
                shared_quant_identity,
                rewriters,
                quant_identity_map,
                align_input_quant_fn,
                align_sign)
        else:
            rewriters.append(InsertModuleCallAfter(shared_quant_identity_name, inp_node))


def _get_quant_module(model, node, quant_identity_map, quant_act_map, unsigned_act_tuple):
    """
    Generate a QuantIdentity node that will be used for requantization step around a node.
    If all inputs to that node are unsigned, the QuantIdentity will also be unsigned, otherwise it
    is signed.
    """
    if are_inputs_unsigned(model, node, [], quant_act_map, unsigned_act_tuple):
        quant_module_class, quant_module_kwargs = quant_identity_map['unsigned']
    else:
        quant_module_class, quant_module_kwargs = quant_identity_map['signed']
    quant_module = quant_module_class(**quant_module_kwargs)
    quant_module_name = node.name + '_quant'
    model.add_module(quant_module_name, quant_module)
    return quant_module, quant_module_name


def residual_handler(
        model, quant_identity_map, quant_act_map, unsigned_act_tuple, align_input_quant_fn):

    def is_converged(model):

        for node in model.graph.nodes:
            if (node.op == 'call_function' and node.target in ADD_FNS + [CAT] or
                    node.op == 'call_method' and node.target in ADD_METHODS):
                rewriters = []
                # If the op is CAT, check that inputs have same sign, and in recursive_input_handler
                # force that the sign is aligned
                same_sign = node.target is CAT

                # If input to the CAT or ADD node are quantized and aligned correctly, continue to
                # the next node
                if are_inputs_quantized_and_aligned(model,
                                                    node, [],
                                                    quant_act_map,
                                                    same_sign=same_sign):
                    continue

                # Generate a QuantIdentity module to use for alignement of the inputs
                shared_quant_identity, shared_quant_identity_name = _get_quant_module(
                    model, node, quant_identity_map, quant_act_map, unsigned_act_tuple)

                # Recursively, for every input node, traverse the graph to determine how to quantize
                # and align that input node.
                recursive_input_handler(
                    model,
                    node,
                    shared_quant_identity_name,
                    shared_quant_identity,
                    rewriters,
                    quant_identity_map,
                    align_input_quant_fn,
                    align_sign=same_sign)
                for rewriter in rewriters:
                    model = rewriter.apply(model)
                model.graph.lint()
                model.recompile()
                return False
        return True

    while not is_converged(model):
        continue

    return model


def add_output_quant_handler(model, quant_identity_map, quant_act_map, unsigned_act_tuple):
    """
    Check the output of every add node, to decide whether it needs to be requantized or not, based
    on the logic of output_quant_handler.
    """
    rewriters = []
    for node in model.graph.nodes:
        if (node.op == 'call_function' and node.target in ADD_FNS or
                node.op == 'call_method' and node.target in ADD_METHODS):
            output_quant_handler(
                model,
                node,
                rewriters,
                is_sign_preserving=True,
                quant_identity_map=quant_identity_map,
                quant_act_map=quant_act_map,
                unsigned_act_tuple=unsigned_act_tuple)
    for rewriter in rewriters:
        model = rewriter.apply(model)
    return model


def layer_handler(
    model,
    layer_map,
    requantize_output,
    quant_identity_map=dict(),
    quant_act_map=dict(),
    unsigned_act_tuple=dict()):
    """
    Replace FP weight layers with their corresponding quantized version
    """
    if requantize_output and (len(quant_identity_map) == 0 or len(quant_act_map) == 0 or
                              len(unsigned_act_tuple) == 0):
        raise RuntimeError("Missing information to requantize output")
    for node in model.graph.nodes:
        rewriters = []
        if node.op == 'call_module':
            module = get_module(model, node.target)
            if isinstance(module, tuple(layer_map.keys())):
                if requantize_output:
                    if len(node.users) > 1 and all(['getitem' in n.name for n in node.users]):
                        for n in node.users:
                            if len(n.users) > 0:
                                output_quant_handler(
                                    model,
                                    n,
                                    rewriters,
                                    is_sign_preserving=isinstance(module, SIGN_PRESERVING_MODULES),
                                    quant_identity_map=quant_identity_map,
                                    quant_act_map=quant_act_map,
                                    unsigned_act_tuple=unsigned_act_tuple)
                    else:
                        output_quant_handler(
                            model,
                            node,
                            rewriters,
                            is_sign_preserving=isinstance(module, SIGN_PRESERVING_MODULES),
                            quant_identity_map=quant_identity_map,
                            quant_act_map=quant_act_map,
                            unsigned_act_tuple=unsigned_act_tuple)
                if layer_map[type(module)] is not None:
                    quant_module_class, quant_module_kwargs = layer_map[type(module)]
                    # Quantize the input if is not quantized, input_quant is not specified,
                    # and the quant_identity_map is provided.
                    # The last requirement is needed to avoid requantizing the input to activations
                    if not are_inputs_quantized_and_aligned(
                            model, node, [], quant_act_map, same_sign=False
                    ) and not 'input_quant' in quant_module_kwargs and len(quant_identity_map) > 0:
                        # Define the source node where to add the requantization step
                        previous_node = node.all_input_nodes[0]
                        # Exclude all the other possible users
                        previous_node_users = list(previous_node.users.keys())
                        previous_node_users.remove(node)

                        act_quant, kwargs_act_quant = quant_identity_map['signed']
                        inp_quant = act_quant(**kwargs_act_quant)
                        name = node.name + '_input_quant'
                        model.add_module(name, inp_quant)
                        rewriter = InsertModuleCallAfter(
                            name, previous_node, tuple(previous_node_users))
                        rewriters.append(rewriter)
                    rewriter = ModuleToModuleByInstance(
                        module, quant_module_class, **quant_module_kwargs)
                    rewriters.append(rewriter)
        for rewriter in rewriters:
            model = rewriter.apply(model)
    return model



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.inject.enum import StatsOp

assert StatsOp

from .stats_op import AbsAve
from .stats_op import AbsMax
from .stats_op import AbsMaxAve
from .stats_op import AbsMaxL2
from .stats_op import AbsMinMax
from .stats_op import AbsPercentile
from .stats_op import L1Norm
from .stats_op import L2Norm
from .stats_op import MeanLearnedSigmaStd
from .stats_op import MeanSigmaStd
from .stats_op import NegativeMinOrZero
from .stats_op import NegativePercentileOrZero
from .stats_op import PercentileInterval
from .stats_wrapper import _ParameterListStats
from .stats_wrapper import _RuntimeStats
from .stats_wrapper import _Stats
from .stats_wrapper import DEFAULT_MOMENTUM
from .stats_wrapper import SCALAR_SHAPE



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from enum import auto

from brevitas.utils.python_utils import AutoName


class BitWidthImplType(AutoName):
    """

    """
    CONST = auto()
    PARAMETER = auto()
    STATEFUL_CONST = auto()


class QuantType(AutoName):
    """

    """
    BINARY = auto()
    TERNARY = auto()
    INT = auto()
    FP = auto()


class RestrictValueType(AutoName):
    """

    """
    FP = auto()
    LOG_FP = auto()
    INT = auto()
    POWER_OF_TWO = auto()


class FloatToIntImplType(AutoName):
    """

    """
    ROUND = auto()
    CEIL = auto()
    FLOOR = auto()
    ROUND_TO_ZERO = auto()
    DPU = auto()
    LEARNED_ROUND = auto()


class LearnedRoundImplType(AutoName):
    """
    """
    HARD_SIGMOID = auto()
    SIGMOID = auto()


class ScalingImplType(AutoName):
    """

    """
    HE = auto()
    CONST = auto()
    STATS = auto()
    AFFINE_STATS = auto()
    PARAMETER = auto()
    PARAMETER_FROM_STATS = auto()


class StatsOp(AutoName):
    """

    """
    # One sided statistics over absolute value
    # Typically adopted for symmetric quantization
    MAX = auto()
    AVE = auto()
    MAX_AVE = auto()
    MEAN_SIGMA_STD = auto()
    MEAN_LEARN_SIGMA_STD = auto()
    PERCENTILE = auto()
    # Two sided statistics
    # Typically adopted for asymmetric quantization
    MIN_MAX = auto()
    PERCENTILE_INTERVAL = auto()



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC

import torch

from brevitas.export.common import to_0dim_if_scalar
from brevitas.export.common import to_item_if_0dim
from brevitas.export.common.handler.base import BaseHandler
from brevitas.export.common.handler.qcdq import DQMixin
from brevitas.export.common.handler.qcdq import QCDQActQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQBiasQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQMixin
from brevitas.export.common.handler.qcdq import QCDQQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQTruncQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQWeightQuantProxyHandlerMixin
from brevitas.proxy import BiasQuantProxyFromInjector


def _itemize_clip_bounds(clip_args):
    if clip_args is not None:
        clip_args['min_val'] = clip_args['min_val'].item()
        clip_args['max_val'] = clip_args['max_val'].item()
    return clip_args


class TorchDQMixin(DQMixin, ABC):

    def dequantize_fn(self, x, scale, zero_point, axis):
        # cast zero_point to float, otherwise if both x
        # and zero_point are uint (as in asym quant)
        # uint - uint can lead to errors. Don't cast x to float
        # as the main float datatype might not be float32 (e.g float16)
        if isinstance(zero_point, torch.Tensor):
            zero_point = zero_point.to(torch.float)
        else:
            zero_point = float(zero_point)
        return (x - zero_point) * scale

    @property
    def flatten_dequantize_params(self):
        return False

    @property
    def itemize_quantize_scalar_params(self):
        return True


class TorchQCDQMixin(TorchDQMixin, QCDQMixin, ABC):

    def __init__(self) -> None:
        super().__init__()
        self.symbolic_kwargs = {}

    @property
    def clip_over_integers(self):
        return True

    @classmethod
    def int8_dtype(cls):
        return torch.qint8

    @classmethod
    def uint8_dtype(cls):
        return torch.quint8

    @classmethod
    def int32_dtype(cls):
        return torch.qint32

    def validate(self, module):
        assert module.bit_width() > 1., 'Binary quant not supported'
        assert module.rounding_mode.upper() == 'ROUND', 'Only round to nearest even supported'

    def quantize_fn(self, x, scale, zero_point, dtype, axis):
        if axis is None:
            y = torch.quantize_per_tensor(x, scale, zero_point, dtype)
        else:
            y = torch.quantize_per_channel(x, scale, zero_point, axis, dtype)
        return y.int_repr()

    def clip_fn(self, x, min_val, max_val):
        return torch.clamp(x, min_val, max_val)

    def forward(self, *args, **kwargs):
        return self.symbolic_execution(*args, **kwargs)


class TorchQCDQHandler(BaseHandler):

    def forward(self, *args, **kwargs):
        return self.symbolic_execution(*args, **kwargs)


class TorchQCDQWeightQuantProxyHandler(TorchQCDQMixin,
                                       QCDQWeightQuantProxyHandlerMixin,
                                       TorchQCDQHandler):

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        clip_args = super().int_clip_symbolic_kwargs(narrow, signed, bit_width)
        return _itemize_clip_bounds(clip_args)


class TorchQCDQActQuantProxyHandler(TorchQCDQMixin, QCDQActQuantProxyHandlerMixin,
                                    TorchQCDQHandler):

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        clip_args = super().int_clip_symbolic_kwargs(narrow, signed, bit_width)
        return _itemize_clip_bounds(clip_args)


class TorchQCDQBiasQuantProxyHandler(TorchDQMixin, QCDQBiasQuantProxyHandlerMixin,
                                     TorchQCDQHandler):
    pass


class TorchQCDQTruncQuantProxyHandler(TorchQCDQMixin,
                                      QCDQTruncQuantProxyHandlerMixin,
                                      TorchQCDQHandler):

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        clip_args = super().int_clip_symbolic_kwargs(narrow, signed, bit_width)
        return _itemize_clip_bounds(clip_args)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.function_wrapper import Identity
from brevitas.core.quant import PrescaledRestrictIntQuant
from brevitas.core.quant import PrescaledRestrictIntQuantWithInputBitWidth
from brevitas.core.quant import RescalingIntQuant
from brevitas.inject import ExtendedInjector
from brevitas.inject import value
from brevitas.inject.enum import QuantType
from brevitas.proxy import BiasQuantProxyFromInjector
from brevitas.quant.solver.common import *
from brevitas.quant.solver.parameter import *

__all__ = [
    'BiasQuantSolver',
    'SolveBiasTensorQuantFromEnum',
    'SolveBiasScalingStatsInputConcatDimFromModule',
    'SolveBiasBitWidthImplFromEnum',
    'SolveBiasScalingPerOutputChannelShapeFromModule']


class SolveBiasScalingStatsInputConcatDimFromModule(ExtendedInjector):
    scaling_stats_input_concat_dim = 0  # bias has only 1 dimension by definition


class SolveBiasScalingPerOutputChannelShapeFromModule(ExtendedInjector):

    @value
    def scaling_per_output_channel_shape(module):
        if isinstance(module, tuple):
            assert all(m.out_channels == module[0].out_channels for m in module)
            module = module[0]
        return (module.out_channels,)


class SolveBiasBitWidthImplFromEnum(ExtendedInjector):

    @value
    def bit_width_impl(bit_width_impl_type, requires_input_bit_width):
        if not requires_input_bit_width:
            return solve_bit_width_impl_from_enum(bit_width_impl_type)
        else:
            return Identity


class SolveBiasTensorQuantFromEnum(SolveIntQuantFromEnum):

    @value
    def tensor_quant(quant_type, requires_input_bit_width, requires_input_scale):
        if quant_type == QuantType.FP:
            return None
        elif quant_type == QuantType.INT:
            if not requires_input_bit_width and requires_input_scale:
                return PrescaledRestrictIntQuant
            elif not requires_input_bit_width and not requires_input_scale:
                return RescalingIntQuant
            else:  # requires_input_bit_width == True
                return PrescaledRestrictIntQuantWithInputBitWidth
        elif quant_type == QuantType.TERNARY:
            raise RuntimeError(f'{quant_type} not supported.')
        elif quant_type == QuantType.BINARY:
            raise RuntimeError(f'{quant_type} not supported.')
        else:
            raise RuntimeError(f'{quant_type} not recognized.')


class BiasQuantSolver(SolveScalingStatsInputViewShapeImplFromEnum,
                      SolveParameterScalingShape,
                      SolveStatsReduceDimFromEnum,
                      SolveScalingStatsOpFromEnum,
                      SolveTensorQuantFloatToIntImplFromEnum,
                      SolveRestrictScalingImplFromEnum,
                      SolveIntScalingImplFromEnum,
                      SolveParameterScalingImplFromEnum,
                      SolveParameterTensorClampImplFromEnum,
                      SolveParameterScalingInitFromEnum,
                      SolveBiasBitWidthImplFromEnum,
                      SolveBiasScalingPerOutputChannelShapeFromModule,
                      SolveBiasScalingStatsInputConcatDimFromModule,
                      SolveBiasTensorQuantFromEnum):
    """
    Translate enum directives to bias-specific quantization core modules.
    It should be placed last in the list of classes a quantizer inherits from,
    to make sure overrides are correctly captured.
    """
    proxy_class = BiasQuantProxyFromInjector



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, Tuple, Union

import torch
from torch import Tensor
from torch.autograd import Function
from torch.nn import Module
from torch.nn import Sequential

from brevitas.export.manager import _set_layer_export_handler
from brevitas.export.manager import _set_layer_export_mode
from brevitas.export.onnx.debug import DebugMarkerFunction
from brevitas.export.onnx.manager import onnx
from brevitas.export.onnx.manager import ONNXBaseManager
from brevitas.quant_tensor import QuantTensor

from .function.acc import TruncAvgPool2dFn
from .function.act import QuantHardTanhFn
from .function.act import QuantReLUFn
from .function.parameter import QuantizedConvNdFn
from .function.parameter import QuantizedLinearFn
from .handler.acc import FINNTruncAvgPool2dHandler
from .handler.act import FINNQuantHardTanhHandler
from .handler.act import FINNQuantIdentityHandler
from .handler.act import FINNQuantReLUHandler
from .handler.parameter import FINNQuantConv1dHandler
from .handler.parameter import FINNQuantConv2dHandler
from .handler.parameter import FINNQuantLinearHandler
from .transform import move_quant_attributes_into_annotations
from .transform import restore_domain
from .utils import finn_datatype


class _InputQuantTensorFunction(Function):
    "Account symbolically for scale and zero-point of an input quant tensor"

    @staticmethod
    def symbolic(g, x, scale, zero_point):
        # x is assumed to be an integer valued tensor here
        if zero_point is not None:
            x = g.op('Sub', x, zero_point)
        if scale is not None:
            x = g.op('Mul', x, scale)
        return x

    @staticmethod
    def forward(ctx, x, scale, zero_point):
        return x


class _InputPreprocessingModule(Module):

    def __init__(self, scale, zero_point):
        super(_InputPreprocessingModule, self).__init__()
        if scale is not None:
            self.register_buffer('scale', scale)
        else:
            self.scale = None
        if zero_point is not None:
            self.register_buffer('zero_point', zero_point)
        else:
            self.zero_point = None

    def forward(self, x):
        if torch.onnx.is_in_onnx_export():
            x = _InputQuantTensorFunction.apply(x, self.scale, self.zero_point)
        return x


def set_quant_tensor_datatype(model, tensor_name, datatype: str):
    qa = onnx.TensorAnnotation()
    dt = onnx.StringStringEntryProto()
    dt.key = "finn_datatype"
    dt.value = datatype
    qa.tensor_name = tensor_name
    qa.quant_parameter_tensor_names.append(dt)
    model.graph.quantization_annotation.append(qa)


class FINNManager(ONNXBaseManager):
    target_name = 'FINN'

    handlers = [
        FINNQuantLinearHandler,
        FINNQuantConv1dHandler,
        FINNQuantConv2dHandler,
        FINNQuantReLUHandler,
        FINNQuantIdentityHandler,
        FINNQuantHardTanhHandler,
        FINNTruncAvgPool2dHandler]

    custom_fns = [
        DebugMarkerFunction,
        QuantizedConvNdFn,
        QuantizedLinearFn,
        QuantReLUFn,
        QuantHardTanhFn,
        TruncAvgPool2dFn]

    model_transforms = [move_quant_attributes_into_annotations, restore_domain]

    onnx_passes = [
        # use initializers instead of Constant nodes for fixed params
        "extract_constant_to_initializer",  # remove unused graph inputs & initializers
        "eliminate_unused_initializer"]

    @classmethod
    def set_export_mode(cls, module: Module, enabled: bool):
        _set_layer_export_mode(module, enabled)

    @classmethod
    def set_export_handler(cls, module: Module):
        _set_layer_export_handler(cls, module)

    @classmethod
    def export(
            cls,
            module: Module,
            args: Optional[Union[Tensor, QuantTensor, Tuple]] = None,
            export_path: Optional[str] = None,
            input_shape: Optional[Tuple[int, ...]] = None,  # legacy syntax, alternative to args
            input_t: Optional[Union[Tensor,
                                    QuantTensor]] = None,  # legacy syntax, alternative to args
            disable_warnings=True,
            **onnx_export_kwargs):
        if ((input_t is not None and isinstance(input_t, QuantTensor) or
             args is not None and isinstance(args, QuantTensor)) and bool(input_t) != bool(args)):

            args = args or input_t  # If either one is not None, args will be not None
            input_t = None  # Keep only args as not None

            if args.is_not_none:
                assert args.is_valid, 'Input QuantTensor is not properly quantized'
            training_state = module.training
            preprocessing_module = _InputPreprocessingModule(args.scale, args.zero_point)
            module = Sequential(preprocessing_module, module)
            module.train(training_state)
        onnx_model = cls.export_onnx(
            module, args, export_path, input_shape, input_t, disable_warnings, **onnx_export_kwargs)
        if args is not None and isinstance(args, QuantTensor):
            bit_width = args.bit_width
            signed = args.signed
            if bit_width is not None and signed is not None:
                # '0' is the name of the input tensor to the model unless otherwise specified
                if 'input_names' in onnx_export_kwargs and onnx_export_kwargs['input_names']:
                    input_name = onnx_export_kwargs['input_names'][0]
                else:
                    input_name = '0'
                set_quant_tensor_datatype(onnx_model, input_name, finn_datatype(bit_width, signed))
                if export_path is not None:
                    onnx.save(onnx_model, export_path)
        return onnx_model



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.quant import *
from brevitas.core.quant import QuantType
from brevitas.inject import ExtendedInjector
from brevitas.inject import this
from brevitas.inject import value
from brevitas.proxy import WeightQuantProxyFromInjector
from brevitas.quant.solver.common import *
from brevitas.quant.solver.parameter import *

__all__ = [
    'SolveWeightTensorQuantFromEnum',
    'SolveWeightScalingStatsInputDimsFromModule',
    'SolveWeightScalingPerOutputChannelShapeFromModule',
    'WeightQuantSolver']


class SolveWeightTensorQuantFromEnum(SolveIntQuantFromEnum):

    @value
    def tensor_quant(quant_type):
        if quant_type == QuantType.FP:
            return None
        elif quant_type == QuantType.INT:
            return RescalingIntQuant
        elif quant_type == QuantType.TERNARY:
            return TernaryQuant
        elif quant_type == QuantType.BINARY:
            return BinaryQuant
        else:
            raise RuntimeError(f'{quant_type} not recognized.')


class SolveWeightScalingPerOutputChannelShapeFromModule(ExtendedInjector):

    @value
    def out_channels(module):
        if isinstance(module, tuple):
            assert all(m.out_channels == module[0].out_channels for m in module)
            module = module[0]
        return module.out_channels

    @value
    def weight_ndims(module):
        if isinstance(module, tuple):
            assert all(len(m.weight.size()) == len(module[0].weight.size()) for m in module)
            module = module[0]
        return len(module.weight.size())

    @value
    def scaling_per_output_channel_shape(weight_ndims, output_channel_dim, out_channels):
        shape = [1] * weight_ndims
        shape[output_channel_dim] = out_channels
        return tuple(shape)


class SolveWeightScalingStatsInputDimsFromModule(ExtendedInjector):

    #  during per-channel quantization weights are always permuted and reshaped first
    #  such that output channels are dim 0 and the remaining features are dim 1,
    #  along which we concatenate
    @value
    def scaling_stats_input_concat_dim(scaling_per_output_channel):
        if scaling_per_output_channel:
            return 1
        else:
            return 0

    @value
    def permute_dims(module, output_channel_dim):
        if output_channel_dim != 0:
            if isinstance(module, tuple):
                assert all(len(m.weight.shape) == len(module[0].weight.shape) for m in module)
                module = module[0]
            dims = list(range(0, len(module.weight.shape)))
            dims[0], dims[output_channel_dim] = dims[output_channel_dim], dims[0]
            return tuple(dims)
        else:
            return None

    @value
    def output_channel_dim(module):
        if isinstance(module, tuple):
            assert all(m.output_channel_dim == module[0].output_channel_dim for m in module)
            module = module[0]
        return module.output_channel_dim


class WeightQuantSolver(SolveWeightScalingStatsInputDimsFromModule,
                        SolveScalingStatsInputViewShapeImplFromEnum,
                        SolveStatsReduceDimFromEnum,
                        SolveScalingStatsOpFromEnum,
                        SolveBitWidthImplFromEnum,
                        SolveTensorQuantFloatToIntImplFromEnum,
                        SolveRestrictScalingImplFromEnum,
                        SolveIntScalingImplFromEnum,
                        SolveParameterScalingImplFromEnum,
                        SolveParameterTensorClampImplFromEnum,
                        SolveParameterScalingInitFromEnum,
                        SolveParameterScalingShape,
                        SolveWeightScalingPerOutputChannelShapeFromModule,
                        SolveWeightTensorQuantFromEnum):
    """
    Translate enum and shape directives to weight-specific quantization core modules.
    It should be placed last in the list of classes a quantizer inherits from,
    to make sure overrides are correctly captured.
    """
    proxy_class = WeightQuantProxyFromInjector



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

# for retrocompatibility

from brevitas.export import enable_debug  # noqa
from brevitas.export import export_brevitas_onnx  # noqa
from brevitas.export import export_finn_onnx  # noqa
from brevitas.export import export_onnx_qcdq  # noqa
from brevitas.export import export_onnx_qop  # noqa
from brevitas.export import export_qonnx  # noqa



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from configparser import ConfigParser
import os

import torch
from torch import hub

__all__ = [
    'cnv_1w1a',
    'cnv_1w2a',
    'cnv_2w2a',
    'sfc_1w1a',
    'sfc_1w2a',
    'sfc_2w2a',
    'tfc_1w1a',
    'tfc_1w2a',
    'tfc_2w2a',
    'lfc_1w1a',
    'lfc_1w2a',
    'resnet18_4w4a',
    'model_with_cfg']

from .CNV import cnv
from .FC import fc
from .resnet import quant_resnet18

model_impl = {'CNV': cnv, 'FC': fc, 'RESNET18': quant_resnet18}


def get_model_cfg(name):
    cfg = ConfigParser()
    current_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(current_dir, '..', 'cfg', name.lower() + '.ini')
    assert os.path.exists(config_path), f"{config_path} not found."
    cfg.read(config_path)
    return cfg


def model_with_cfg(name, pretrained):
    cfg = get_model_cfg(name)
    arch = cfg.get('MODEL', 'ARCH')
    model = model_impl[arch](cfg)
    if pretrained:
        checkpoint = cfg.get('MODEL', 'PRETRAINED_URL')
        state_dict = hub.load_state_dict_from_url(checkpoint, progress=True, map_location='cpu')
        model.load_state_dict(state_dict, strict=True)
    return model, cfg


def cnv_1w1a(pretrained=True):
    model, _ = model_with_cfg('cnv_1w1a', pretrained)
    return model


def cnv_1w2a(pretrained=True):
    model, _ = model_with_cfg('cnv_1w2a', pretrained)
    return model


def cnv_2w2a(pretrained=True):
    model, _ = model_with_cfg('cnv_2w2a', pretrained)
    return model


def sfc_1w1a(pretrained=True):
    model, _ = model_with_cfg('sfc_1w1a', pretrained)
    return model


def sfc_1w2a(pretrained=True):
    model, _ = model_with_cfg('sfc_1w2a', pretrained)
    return model


def sfc_2w2a(pretrained=True):
    model, _ = model_with_cfg('sfc_2w2a', pretrained)
    return model


def tfc_1w1a(pretrained=True):
    model, _ = model_with_cfg('tfc_1w1a', pretrained)
    return model


def tfc_1w2a(pretrained=True):
    model, _ = model_with_cfg('tfc_1w2a', pretrained)
    return model


def tfc_2w2a(pretrained=True):
    model, _ = model_with_cfg('tfc_2w2a', pretrained)
    return model


def lfc_1w1a(pretrained=True):
    model, _ = model_with_cfg('lfc_1w1a', pretrained)
    return model


def lfc_1w2a(pretrained=True):
    model, _ = model_with_cfg('lfc_1w2a', pretrained)
    return model


def resnet18_4w4a(pretrained=True):
    model, _ = model_with_cfg('resnet18_4w4a', pretrained)
    return model



from abc import ABC

import torch
from torch import Tensor

from brevitas.export.onnx.standard.function import DequantizeLinearFn
from brevitas.export.onnx.standard.function import IntClipFn
from brevitas.export.onnx.standard.function import QuantizeLinearFn
from brevitas.nn import QuantHardTanh
from brevitas.nn import QuantIdentity
from brevitas.nn import QuantReLU
from brevitas.nn import QuantSigmoid
from brevitas.nn import QuantTanh
from brevitas.nn.quant_layer import QuantNonLinearActLayer as QuantNLAL

from .base import StdQOpONNXQuantLayerHandler


class StdQOpONNXQuantNLALHandler(StdQOpONNXQuantLayerHandler, ABC):

    @classmethod
    def validate(cls, module: QuantNLAL):
        if cls.input_quant_supported and module.is_input_quant_enabled:
            assert not module.is_quant_input_narrow_range, "Narrow range quant not supported."
        elif not cls.input_quant_supported and module.is_input_quant_enabled:
            raise RuntimeError("Input quant not supported.")
        if module.is_act_quant_enabled:
            assert not module.is_quant_act_narrow_range, "Narrow range quant not supported."
        input_bit_width = module.quant_input_bit_width()
        act_bit_width = module.quant_act_bit_width()
        if input_bit_width is not None:
            cls.validate_8b_bit_width(input_bit_width, le_then=True)
        if act_bit_width is not None:
            cls.validate_8b_bit_width(act_bit_width, le_then=True)

    def prepare_for_export(self, module: QuantNLAL):
        self.validate(module)
        if not module.return_quant_tensor:
            output_dequant_symbolic_kwargs = self.output_dequant_symbolic_kwargs(module)
        else:
            output_dequant_symbolic_kwargs = None
        output_quant_symbolic_kwargs = self.output_quant_symbolic_kwargs(module)
        output_clip_symbolic_kwargs = self.output_clip_symbolic_kwargs(module)
        input_dequant_symbolic_kwargs = self.input_dequant_symbolic_kwargs(module)
        input_quant_symbolic_kwargs = self.input_quant_symbolic_kwargs(module)
        input_clip_symbolic_kwargs = self.input_clip_symbolic_kwargs(module)
        if input_quant_symbolic_kwargs is not None:
            input_redequant_symbolic_kwargs = {
                'input_scale': input_quant_symbolic_kwargs['output_scale'],
                'input_zero_point': input_quant_symbolic_kwargs['output_zero_point'],
                'input_axis': input_quant_symbolic_kwargs['input_axis']}
        else:
            input_redequant_symbolic_kwargs = None

        self.symbolic_kwargs = {
            'input_quant_symbolic_kwargs': input_quant_symbolic_kwargs,
            'input_clip_symbolic_kwargs': input_clip_symbolic_kwargs,
            'input_dequant_symbolic_kwargs': input_dequant_symbolic_kwargs,
            'input_redequant_symbolic_kwargs': input_redequant_symbolic_kwargs,
            'output_quant_symbolic_kwargs': output_quant_symbolic_kwargs,
            'output_clip_symbolic_kwargs': output_clip_symbolic_kwargs,
            'output_dequant_symbolic_kwargs': output_dequant_symbolic_kwargs}

    def input_symbolic_execution(self, inp: Tensor):
        input_quant_symbolic_kwargs = self.symbolic_kwargs['input_quant_symbolic_kwargs']
        input_clip_symbolic_kwargs = self.symbolic_kwargs['input_clip_symbolic_kwargs']
        input_dequant_symbolic_kwargs = self.symbolic_kwargs['input_dequant_symbolic_kwargs']
        input_redequant_symbolic_kwargs = self.symbolic_kwargs['input_redequant_symbolic_kwargs']
        if input_dequant_symbolic_kwargs is not None:
            inp = DequantizeLinearFn.apply(inp, *input_dequant_symbolic_kwargs.values())
        if input_quant_symbolic_kwargs is not None:
            inp = QuantizeLinearFn.apply(inp, *input_quant_symbolic_kwargs.values())
            inp = DequantizeLinearFn.apply(inp, *input_redequant_symbolic_kwargs.values())
            if input_clip_symbolic_kwargs is not None:
                inp = IntClipFn.apply(inp, *input_clip_symbolic_kwargs.values())
        return inp

    def output_symbolic_execution(self, out: Tensor):
        output_quant_symbolic_kwargs = self.symbolic_kwargs['output_quant_symbolic_kwargs']
        output_dequant_symbolic_kwargs = self.symbolic_kwargs['output_dequant_symbolic_kwargs']
        output_clip_symbolic_kwargs = self.symbolic_kwargs['output_clip_symbolic_kwargs']
        if output_quant_symbolic_kwargs is not None:
            out = QuantizeLinearFn.apply(out, *output_quant_symbolic_kwargs.values())
            if output_clip_symbolic_kwargs is not None:
                out = IntClipFn.apply(out, *output_clip_symbolic_kwargs.values())
        if output_dequant_symbolic_kwargs is not None:
            out = DequantizeLinearFn.apply(out, *output_dequant_symbolic_kwargs.values())
        return out


class StdQOpONNXQuantReLUHandler(StdQOpONNXQuantNLALHandler):
    handled_layer = QuantReLU
    input_quant_supported = True

    def op_symbolic_execution(self, inp: Tensor):
        return torch.relu(inp)


class StdQOpONNXQuantTanhHandler(StdQOpONNXQuantNLALHandler):
    handled_layer = QuantTanh
    input_quant_supported = True

    def op_symbolic_execution(self, inp: Tensor):
        return torch.tanh(inp)


class StdQOpONNXQuantSigmoidHandler(StdQOpONNXQuantNLALHandler):
    handled_layer = QuantSigmoid
    input_quant_supported = True

    def op_symbolic_execution(self, inp: Tensor):
        return torch.sigmoid(inp)


class StdQOpONNXQuantIdentityHandler(StdQOpONNXQuantNLALHandler):
    handled_layer = QuantIdentity
    input_quant_supported = False

    def op_symbolic_execution(self, inp: Tensor):
        return inp


class StdQOpONNXQuantHardTanhHandler(StdQOpONNXQuantIdentityHandler):
    handled_layer = QuantHardTanh



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from functools import partial
import sys

from torch import nn
import torch.nn.functional as F

from brevitas.nn import QuantHardTanh
from brevitas.nn import QuantLinear
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from brevitas.nn.utils import compute_channel_view_shape
from brevitas.proxy.parameter_quant import BiasQuantProxyFromInjector
from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjector
from brevitas.proxy.runtime_quant import ActQuantProxyFromInjector
from brevitas.proxy.runtime_quant import ClampQuantProxyFromInjector
from brevitas.proxy.runtime_quant import TruncQuantProxyFromInjector
from brevitas.quant_tensor import QuantTensor

from .base import Transform

__all__ = [
    'ClipFloatWeights', 'DisableEnableQuantization', 'bias_correction_mode', 'calibration_mode']

_PARAM_PROXIES = (WeightQuantProxyFromInjector, BiasQuantProxyFromInjector)

_ACC_PROXIES = (TruncQuantProxyFromInjector, ClampQuantProxyFromInjector)

_LAYERS_TO_CLIP = (
    nn.Conv1d,
    nn.Conv2d,
    nn.Conv3d,
    nn.Linear,
    nn.ConvTranspose1d,
    nn.ConvTranspose2d,
    nn.ConvTranspose3d)


def extend_collect_stats_steps(module):
    if hasattr(module, 'collect_stats_steps'):
        # We extend the collect steps in PTQ to match potentially long calibrations
        module.collect_stats_steps = sys.maxsize


def finalize_collect_stats(module):
    if hasattr(module, 'collect_stats_steps') and hasattr(module, 'counter'):
        # If the counter has already reached collect_stats_steps, we do not want to reset it
        # otherwise the restrict_preprocess might be applied twice: during calibration
        # (that happens in training mode) and then when the model is evaluated
        module.counter = max(module.collect_stats_steps, module.counter)


class calibration_mode:

    def __init__(self, model, enabled=True):
        self.model = model
        self.previous_training_state = model.training
        self.disable_quant_inference = DisableEnableQuantization()
        self.enabled = enabled

    def __enter__(self):
        if self.enabled:
            self.model.apply(extend_collect_stats_steps)
            self.disable_quant_inference.apply(
                self.model, is_training=True, quantization_enabled=False)

    def __exit__(self, type, value, traceback):
        self.model.apply(finalize_collect_stats)
        self.disable_quant_inference.apply(
            self.model, is_training=self.previous_training_state, quantization_enabled=True)


class bias_correction_mode:

    def __init__(self, model, enabled=True):
        self.model = model
        self.bias_correction = _BiasCorrection()
        self.enabled = enabled
        self.hooks = []

    def __enter__(self):
        if self.enabled:
            self.bias_correction.register_hook_to_wbiol(self.model, self.hooks)

    def __exit__(self, type, value, traceback):
        self.bias_correction.apply_correction(self.model)
        for hook in self.hooks:
            hook.remove()


class ClipFloatWeights(Transform):

    def __init__(self, threshold=15., layers_to_clip=_LAYERS_TO_CLIP) -> None:
        super(ClipFloatWeights, self).__init__()
        self.threshold = threshold
        self.layers_to_clip = layers_to_clip

    def apply(self, model):
        for module in model.modules():
            if isinstance(module, self.layers_to_clip):
                module.weight.data.clamp_(-self.threshold, self.threshold)
        return model


class DisableEnableQuantization(Transform):

    def __init__(self):
        super(DisableEnableQuantization, self).__init__()
        self.disable_act_quant_hooks = []

    def unpack_input(self, inp):
        if isinstance(inp, tuple):
            inp = inp[0]
        if isinstance(inp, QuantTensor):
            inp = inp.value
        return inp

    def disable_act_quant_hook(self, module, inp, output):
        inp = self.unpack_input(inp)
        if module.fused_activation_quant_proxy is not None:
            inp = module.fused_activation_quant_proxy.activation_impl(inp)
        # consider the first module as representative for the activation fn
        # as this is what would happen with a shared act_quant
        # this gets called both during (empty) input_quant and act_quant
        # but for HardTanh it's not an issue
        if isinstance(module.tracked_module_list[0], QuantHardTanh):
            inp = F.hardtanh(
                inp, min_val=module.quant_injector.min_val, max_val=module.quant_injector.max_val)
        return QuantTensor(value=inp, training=module.training)

    def disable_act_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, ActQuantProxyFromInjector):
                hook = module.register_forward_hook(self.disable_act_quant_hook)
                module.train(is_training)
                self.disable_act_quant_hooks.append(hook)
            elif isinstance(module, _ACC_PROXIES):
                module.train(is_training)
                module.disable_quant = True

    def disable_param_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, _PARAM_PROXIES):
                module.train(is_training)
                module.disable_quant = True

    def enable_act_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, _ACC_PROXIES):
                module.train(is_training)
                module.disable_quant = False
            elif isinstance(module, ActQuantProxyFromInjector):
                module.train(is_training)
        for hook in self.disable_act_quant_hooks:
            hook.remove()
        self.disable_act_quant_hooks = []

    def enable_param_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, _PARAM_PROXIES):
                module.disable_quant = False
                module.train(is_training)

    def apply(self, model, is_training, quantization_enabled):
        if not quantization_enabled:
            self.disable_act_quantization(model, is_training)
            self.disable_param_quantization(model, is_training)
        else:
            self.enable_act_quantization(model, is_training)
            self.enable_param_quantization(model, is_training)


class _BiasCorrection(DisableEnableQuantization):

    LAYERS = (QuantWBIOL,)

    def __init__(self, layers=LAYERS):
        super(_BiasCorrection, self).__init__()
        self.layers = layers
        self.iterations = {}
        self.correction_map = {}
        self.float_mean_map = {}
        self.collect_float_mean_hooks = []
        self.correct_bias_hooks = []

    def compute_mean(self, inp, transpose_dim):
        inp = inp.transpose(0, transpose_dim)
        return inp.reshape(inp.shape[0], -1).mean(dim=1).detach()

    def channel_dim(self, inp, module):
        if len(inp.shape) == 3 and isinstance(module, QuantLinear):
            channel_dim = 2
        else:
            channel_dim = 1
        return channel_dim

    def collect_float_mean_hook(self, module, inp, name, parent_module):
        inp = self.unpack_input(inp)
        if name in self.float_mean_map.keys():
            raise RuntimeError("Module to bias-correct called multiple times, not supported.")
        transpose_dim = self.channel_dim(inp, parent_module)
        self.float_mean_map[name] = self.compute_mean(inp, transpose_dim)

    def update_correction(self, name, error):
        if name not in self.correction_map:
            self.correction_map[name] = error
        else:
            self.correction_map[name] += error

    def apply_correction(self, model):
        for name, module in model.named_modules():
            if name in self.correction_map.keys():
                correction = self.correction_map[name] / self.iterations[name]
                if module.bias is not None:
                    module.bias.data += correction
                else:
                    module.register_parameter(
                        'bias', nn.Parameter(correction).to(module.weight.device))

    def correct_bias_hook(self, module, inp, name, parent_module):
        inp = self.unpack_input(inp)
        if name in self.float_mean_map.keys():
            transpose_dim = self.channel_dim(inp, parent_module)
            quant_mean = self.compute_mean(inp, transpose_dim)
            error = self.float_mean_map[name] - quant_mean
            self.update_correction(name, error)
            del self.float_mean_map[name]
            inp_broadcast_shape = compute_channel_view_shape(
                inp, channel_dim=self.channel_dim(inp, parent_module))
            return inp + error.reshape(inp_broadcast_shape)

    def register_collect_float_mean_hook(self, module, name):
        hook_fn = partial(self.collect_float_mean_hook, name=name, parent_module=module)
        hook = module.output_quant.register_forward_pre_hook(hook_fn)
        self.collect_float_mean_hooks.append(hook)

    def register_correct_bias_hook(self, module, name):
        hook_fn = partial(self.correct_bias_hook, name=name, parent_module=module)
        hook = module.output_quant.register_forward_pre_hook(hook_fn)
        self.correct_bias_hooks.append(hook)

    def float_mean_hooks_cleanup(self):
        for hook in self.collect_float_mean_hooks:
            hook.remove()
        self.collect_float_mean_hooks = []

    def correct_bias_hooks_cleanup(self):
        for hook in self.correct_bias_hooks:
            hook.remove()
        self.correct_bias_hooks = []

    def register_hook_to_wbiol(self, model, hooks):
        """
        Forward hooks are registered to the WBIOL layers.
        In this way, if more hooks are registered to these layers, they will be only called once per module call.
        We perform two more forwards for each WBIOL, but since we call forward and not __call__, eventual hooks would be skipped.
        This is a desired behaviour, since these extra forwards are only necessary to compute the bias correction factor.

        If we registered a single hook for the entire model, and hooks were added to WBIOL layers,
        these would be called by our "extra forwards", which would be an unexpected behaviours.
        """
        for name, module in model.named_modules():
            if isinstance(module, self.layers):
                self.iterations[name] = 0
                hook_fn = partial(self.forward_hook_wbiol, name=name)
                hooks.append(module.register_forward_hook(hook_fn))

    def forward_hook_wbiol(self, module, inp, output, name):
        """
        After each forward, we perform two extra forwards to register FP and Quant outputs
        and the error between the two.

        We do not return the original Quant output, but its "corrected version", i.e., the FP version
        """
        self.disable_act_quantization(module, is_training=False)
        self.disable_param_quantization(module, is_training=False)
        self.register_collect_float_mean_hook(module, name)
        module.forward(*inp)  # Required to avoid infinite recursion
        self.float_mean_hooks_cleanup()
        self.enable_act_quantization(module, is_training=False)
        self.enable_param_quantization(module, is_training=False)
        self.register_correct_bias_hook(module, name)
        out = module.forward(*inp)  # Required to avoid infinite recursion
        self.iterations[name] += 1
        return out



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from dependencies import this
from dependencies import value

from brevitas.core.bit_width import BitWidthConst
from brevitas.core.bit_width import BitWidthStatefulConst
from brevitas.core.function_wrapper import OverOutputChannelView
from brevitas.core.function_wrapper import RoundToZeroSte
from brevitas.core.function_wrapper import TensorClamp
from brevitas.core.function_wrapper import TensorClampSte
from brevitas.core.function_wrapper.ops_ste import CeilSte
from brevitas.core.quant import ClampedBinaryQuant
from brevitas.core.quant.int import DecoupledRescalingIntQuant
from brevitas.core.quant.int import DecoupledRescalingIntQuantWithInput
from brevitas.core.quant.int_base import DecoupledIntQuant
from brevitas.core.restrict_val import FloatRestrictValue
from brevitas.core.restrict_val import LogFloatRestrictValue
from brevitas.core.scaling import AccumulatorAwareParameterPreScaling
from brevitas.core.scaling import IntScaling
from brevitas.core.scaling import ParameterPreScalingWeightNorm
from brevitas.core.scaling import ParameterScaling
from brevitas.core.scaling import SCALAR_SHAPE
from brevitas.core.scaling import SCALING_STATS_REDUCE_DIM
from brevitas.core.scaling import StatsFromParameterScaling
from brevitas.core.stats import AbsMax
from brevitas.core.stats import AbsMaxL2
from brevitas.core.stats import L1Norm
from brevitas.core.stats import L2Norm
from brevitas.core.stats import NegativeMinOrZero
from brevitas.core.stats import NegativePercentileOrZero
from brevitas.core.utils import SingleArgStatelessBuffer
from brevitas.core.zero_point import ParameterFromRuntimeZeroPoint
from brevitas.core.zero_point import StatsFromParameterZeroPoint
from brevitas.core.zero_point import ZeroZeroPoint
from brevitas.inject import ExtendedInjector
from brevitas.inject.enum import BitWidthImplType
from brevitas.inject.enum import FloatToIntImplType
from brevitas.inject.enum import QuantType
from brevitas.inject.enum import RestrictValueType
from brevitas.inject.enum import ScalingImplType
from brevitas.inject.enum import StatsOp
from brevitas.proxy import DecoupledWeightQuantProxyFromInjector
from brevitas.proxy import DecoupledWeightQuantWithInputProxyFromInjector
from brevitas.quant.solver.parameter import ParameterFromStatsScalingInit
from brevitas.quant.solver.parameter import SolveParameterScalingShape
from brevitas.quant.solver.weight import SolveWeightScalingPerOutputChannelShapeFromModule
from brevitas.quant.solver.weight import SolveWeightScalingStatsInputDimsFromModule

__all__ = [
    'MaxStatsScaling',
    'MinMaxStatsScaling',
    'ParamFromRuntimePercentileScaling',
    'ParamFromRuntimePercentileIntervalScaling',
    'ParamFromRuntimeMinMaxScaling',
    'ParamMinMaxInitScaling',
    'IntQuant',
    'NarrowIntQuant',
    'UintQuant',
    'ShiftedMinUintQuant',
    'ShiftedParamFromPercentileUintQuant',
    'PerChannelFloatScaling8bit',
    'PerTensorFloatScaling8bit',
    'PerTensorPoTScaling8bit',
    'SignedBinaryClampedConst',
    'WeightPerTensorFloatDecoupledL2Param',
    'WeightPerChannelFloatDecoupled',
    'WeightNormPerChannelFloatDecoupled',
    'BatchQuantStatsScaling1d',
    'BatchQuantStatsScaling2d',
    'AccumulatorAwareWeightQuant']


class MaxStatsScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.STATS
    scaling_stats_op = StatsOp.MAX
    scaling_min_val = 1e-10


class MinMaxStatsScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.STATS
    scaling_stats_op = StatsOp.MIN_MAX
    scaling_min_val = 1e-10


class ParamFromRuntimePercentileScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    scaling_stats_op = StatsOp.PERCENTILE
    high_percentile_q = 99.999
    collect_stats_steps = 300
    scaling_min_val = 1e-10


class BatchQuantStatsScaling1d(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.AFFINE_STATS
    scaling_stats_op = StatsOp.MAX_AVE
    scaling_stats_momentum = None  # perform a true average rather than exp average
    affine_shift_scale = False  # max_ave statistics are scaled but not shifted
    scaling_stats_permute_dims = (1, 0, 2)  # assuming (N, C, F), put channel dimension first
    scaling_min_val = 1e-10


class BatchQuantStatsScaling2d(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.AFFINE_STATS
    scaling_stats_op = StatsOp.MAX_AVE
    scaling_stats_momentum = None  # perform a true average rather than exp average
    affine_shift_scale = False  # max_ave statistics are scaled but not shifted
    scaling_stats_permute_dims = (1, 0, 2, 3)  # assuming (N, C, H, W), put channel dimension first
    scaling_min_val = 1e-10


class ParamFromRuntimeMinMaxScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    scaling_stats_op = StatsOp.MIN_MAX
    collect_stats_steps = 300
    scaling_min_val = 1e-10


class ParamFromRuntimePercentileIntervalScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    scaling_stats_op = StatsOp.PERCENTILE_INTERVAL
    high_percentile_q = 99.999
    low_percentile_q = 0.001
    collect_stats_steps = 300
    scaling_min_val = 1e-10


class ParamMinMaxInitScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.PARAMETER


class IntQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = False
    signed = True
    zero_point_impl = ZeroZeroPoint


class NarrowIntQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = True
    signed = True
    zero_point_impl = ZeroZeroPoint


class UintQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = False
    signed = False
    zero_point_impl = ZeroZeroPoint


class ShiftedMinUintQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = False
    signed = False
    quantize_zero_point = True
    zero_point_impl = StatsFromParameterZeroPoint
    zero_point_stats_impl = NegativeMinOrZero
    zero_point_shape = this.scaling_shape
    zero_point_stats_input_view_shape_impl = this.scaling_stats_input_view_shape_impl
    zero_point_stats_input_concat_dim = this.scaling_stats_input_concat_dim


class ShiftedParamFromPercentileUintQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = False
    signed = False
    quantize_zero_point = True
    zero_point_impl = ParameterFromRuntimeZeroPoint
    zero_point_stats_impl = NegativePercentileOrZero
    low_percentile_q = 0.001
    zero_point_shape = this.scaling_shape
    zero_point_stats_input_view_shape_impl = this.scaling_stats_input_view_shape_impl


class PerChannelFloatScaling8bit(ExtendedInjector):
    """
    """
    scaling_per_output_channel = True
    restrict_scaling_type = RestrictValueType.FP
    bit_width = 8


class PerTensorFloatScaling8bit(ExtendedInjector):
    """
    """
    scaling_per_output_channel = False
    restrict_scaling_type = RestrictValueType.FP
    bit_width = 8


class PerTensorPoTScaling8bit(ExtendedInjector):
    """
    """
    scaling_per_output_channel = False
    restrict_scaling_type = RestrictValueType.POWER_OF_TWO
    bit_width = 8
    restrict_value_float_to_int_impl = CeilSte


class SignedBinaryClampedConst(ExtendedInjector):
    tensor_quant = ClampedBinaryQuant
    scaling_impl_type = ScalingImplType.CONST
    restrict_scaling_type = RestrictValueType.FP
    float_to_int_impl_type = FloatToIntImplType.ROUND
    scaling_per_output_channel = False
    narrow_range = True
    signed = True


class PerTensorConstScaling2bit(ExtendedInjector):
    scaling_impl_type = ScalingImplType.CONST
    restrict_scaling_type = RestrictValueType.FP
    scaling_per_output_channel = False
    bit_width = 2


class WeightPerTensorFloatDecoupledL2Param(SolveWeightScalingStatsInputDimsFromModule):
    """
    Experimental narrow per-tensor signed int weight quantizer fragment with decoupled L2,inf
    normalization and learned scaling.
    """

    @value
    def scaling_init(scaling_init_impl):
        return scaling_init_impl()

    proxy_class = DecoupledWeightQuantProxyFromInjector
    tensor_quant = DecoupledRescalingIntQuant
    decoupled_int_quant = DecoupledIntQuant
    tensor_clamp_impl = TensorClampSte
    pre_scaling_impl = StatsFromParameterScaling
    scaling_stats_impl = AbsMaxL2
    scaling_stats_input_view_shape_impl = OverOutputChannelView
    scaling_stats_input_concat_dim = 0
    stats_reduce_dim = SCALING_STATS_REDUCE_DIM
    restrict_scaling_impl = FloatRestrictValue
    scaling_shape = SCALAR_SHAPE
    scaling_impl = ParameterScaling
    scaling_init_impl = ParameterFromStatsScalingInit
    parameter_stats_scaling_init_impl = this.pre_scaling_impl
    int_scaling_impl = IntScaling
    zero_point_impl = ZeroZeroPoint
    pre_zero_point_impl = ZeroZeroPoint
    bit_width_impl = BitWidthConst
    narrow_range = True
    signed = True


class WeightPerChannelFloatDecoupled(SolveWeightScalingStatsInputDimsFromModule,
                                     SolveWeightScalingPerOutputChannelShapeFromModule,
                                     SolveParameterScalingShape):
    """
    Experimental narrow per-channel signed int weight quantizer fragment with decoupled Linf
    normalization and learned scaling.
    """

    @value
    def scaling_init(scaling_init_impl):
        return scaling_init_impl()

    proxy_class = DecoupledWeightQuantProxyFromInjector
    tensor_quant = DecoupledRescalingIntQuant
    decoupled_int_quant = DecoupledIntQuant
    tensor_clamp_impl = TensorClampSte
    pre_scaling_impl = StatsFromParameterScaling
    scaling_stats_impl = AbsMax
    restrict_scaling_impl = FloatRestrictValue
    scaling_impl = ParameterScaling
    scaling_init_impl = ParameterFromStatsScalingInit
    parameter_stats_scaling_init_impl = this.pre_scaling_impl
    int_scaling_impl = IntScaling
    zero_point_impl = ZeroZeroPoint
    pre_zero_point_impl = ZeroZeroPoint
    bit_width_impl = BitWidthConst
    narrow_range = True
    signed = True
    scaling_stats_input_view_shape_impl = OverOutputChannelView
    stats_reduce_dim = SCALING_STATS_REDUCE_DIM
    scaling_per_output_channel = True


class WeightNormPerChannelFloatDecoupled(SolveWeightScalingStatsInputDimsFromModule,
                                         SolveWeightScalingPerOutputChannelShapeFromModule,
                                         SolveParameterScalingShape):
    """Experimental narrow per-channel weight normalization-based signed integer quantizer
    based on `Quantized Neural Networks for Low-Precision Accumulation with Guaranteed
    Overflow Avoidance` by I. Colbert, A. Pappalardo, and J. Petri-Koenig.

    The formulation for weight normalization-based quantization is given below:
        `y = clip(round( (g / s) * (w / norm(w)) )) * s`

    The default quantizer uses the decoupled rescaling integer quantization arithmetic
    where the weight normalization calculation and parameterization are combined with the
    scaling factor to become the pre-clipping scaling factor (i.e., `pre_scale`) and the
    scaling factor is the post-clipping scaling factor (i.e., `post_scale`). For further
    details on the arithmetic, see `ParameterPreScalingWeightNorm`. For further details
    on the weight normalization-based quantization technique, see the referenced paper."""

    @value
    def scaling_init(scaling_init_impl):
        return scaling_init_impl()

    proxy_class = DecoupledWeightQuantProxyFromInjector
    tensor_quant = DecoupledRescalingIntQuant
    decoupled_int_quant = DecoupledIntQuant
    tensor_clamp_impl = TensorClamp
    scaling_impl = ParameterScaling
    restrict_scaling_impl = FloatRestrictValue
    scaling_stats_impl = AbsMax
    scaling_init_impl = ParameterFromStatsScalingInit
    parameter_stats_scaling_init_impl = StatsFromParameterScaling
    pre_scaling_impl = ParameterPreScalingWeightNorm
    restrict_pre_scaling_impl = LogFloatRestrictValue
    normalize_stats_impl = L2Norm
    pre_scaling_shape = this.scaling_shape  # TODO: decouple pre_scaling_shape from scaling_shape
    int_scaling_impl = SingleArgStatelessBuffer(1.)
    zero_point_impl = ZeroZeroPoint
    pre_zero_point_impl = ZeroZeroPoint
    bit_width_impl = BitWidthConst
    narrow_range = True
    signed = True
    scaling_stats_input_view_shape_impl = OverOutputChannelView
    stats_reduce_dim = SCALING_STATS_REDUCE_DIM
    scaling_per_output_channel = True


class AccumulatorAwareWeightQuant(WeightNormPerChannelFloatDecoupled):
    """Experimental accumulator-aware weight quantizer based on `Quantized Neural Networks
    for Low-Precision Accumulation with Guaranteed Overflow Avoidance` by I. Colbert,
    A. Pappalardo, and J. Petri-Koenig.

    The formulation is based on weight normalization-based quantization as given below:
        `y = clip(round( (g / s) * (w / norm(w)) )) * s`
    where `g` is the constrained such that the upper-bound on the l1-norm of the quantized
    weights satisfies the bounds derived in the referenced paper.

    The default quantizer uses input-aware decoupled rescaling integer quantization arithmetic
    where the constrained weight normalization calculation and parameterization are combined
    with the scaling factor to become the pre-clipping scaling factor (i.e., `pre_scale`) an
    the scaling factor is the post-clipping scaling factor (i.e., `post_scale`). For further
    details on the arithmetic, see `AccumulatorAwareParameterPreScalingWeightNorm`. For further
    details on accumulator-aware quantization (A2Q) technique, see the referenced paper."""

    @value
    def accumulator_bit_width_impl(accumulator_bit_width):
        return BitWidthStatefulConst(accumulator_bit_width)

    proxy_class = DecoupledWeightQuantWithInputProxyFromInjector
    tensor_quant = DecoupledRescalingIntQuantWithInput
    pre_scaling_impl = AccumulatorAwareParameterPreScaling
    pre_scaling_min_val = 1e-8
    accumulator_bit_width = 32  # default maximum accumulator width is 32 bits
    normalize_stats_impl = L1Norm  # required to align with derivations in paper
    float_to_int_impl = RoundToZeroSte  # required to ensure no upwards rounding violates constraints



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

# Fix mnist download on older torchvision versions
from urllib import request

from .models import *

opener = request.build_opener()
opener.addheaders = [('User-agent', 'Mozilla/5.0')]
request.install_opener(opener)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

try:
    import torch.nn.quantized.functional as qF

# Skip for pytorch 1.1.0
except:
    qF = None



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .clamp import ClampMin
from .clamp import ScalarClamp
from .clamp import TensorClamp
from .misc import Identity
from .misc import InplaceLogTwo
from .misc import LogTwo
from .misc import PowerOfTwo
from .ops_ste import CeilSte
from .ops_ste import DPURoundSte
from .ops_ste import FloorSte
from .ops_ste import InplaceTensorClampSte
from .ops_ste import RoundSte
from .ops_ste import RoundToZeroSte
from .ops_ste import ScalarClampMinSte
from .ops_ste import TensorClampSte
from .shape import OverBatchOverOutputChannelView
from .shape import OverBatchOverTensorView
from .shape import OverOutputChannelView
from .shape import OverTensorView
from .shape import StatsInputViewShapeImpl



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import copy

from torch.nn import Sequential


class TupleSequential(Sequential):

    def output(self, mod, input):
        if isinstance(input, tuple):
            return mod(*input)
        else:
            return mod(input)

    def forward(self, *input):
        modules = list(self._modules.values())
        out = self.output(modules[0], input)
        for mod in modules[1:]:
            out = self.output(mod, out)
        return out


def torch_partial_deepcopy(model):
    """
    Performs a deepcopy of a torch.nn.Module, except for all the parameters that are instead passed by reference
    """
    memo = {}
    for p in model.parameters():
        memo[id(p)] = copy.copy(p)  # Shallow copy of parameters
    model_copy = copy.deepcopy(model, memo)
    return model_copy



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import math
from typing import Optional, Tuple

import torch
from torch import Tensor
from torch.nn import Parameter

import brevitas
from brevitas import config
from brevitas.core.utils import StatelessBuffer
from brevitas.function.ops import max_int

from .stats_wrapper import SCALAR_SHAPE

DEFAULT_STD_DEV_EPSILON = 1e-8


class NegativeMinOrZero(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(NegativeMinOrZero, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.zero = StatelessBuffer(torch.tensor(0.0))

    @brevitas.jit.script_method
    def forward(self, x: Tensor) -> Tensor:
        if self.stats_reduce_dim is None:
            min_val = torch.min(x)
        else:
            min_val = torch.min(x, dim=self.stats_reduce_dim)[0]
        min_val = torch.where(
            min_val <= self.zero().to(min_val.dtype), min_val, self.zero().to(min_val.dtype))
        return min_val


class AbsPercentile(brevitas.jit.ScriptModule):
    __constants__ = ['q', 'stats_reduce_dim']

    def __init__(
            self, high_percentile_q: float, stats_reduce_dim: Optional[int], percentile_q=None):
        super(AbsPercentile, self).__init__()
        if percentile_q is not None:
            raise RuntimeError("percentile_q is deprecated, please pass high_percentile_q.")
        assert high_percentile_q <= 100, "q has to be a percentage"
        self.q = high_percentile_q
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            # k is 1-indexed, so round away from zero
            k = int(math.floor(.01 * self.q * x.numel() + 0.5))
            result = x.abs().view(-1).kthvalue(k).values
        else:
            # assuming x is two dimensional, get the other dimension
            assert len(x.size()) == 2, "Only 2-dim input is supported."
            other_dim = abs(self.stats_reduce_dim - 1)
            dim_slice = torch.narrow(x, dim=other_dim, start=0, length=1)
            # k is 1-indexed, so round away from zero
            k = int(math.floor(.01 * self.q * dim_slice.numel() + 0.5))
            result = x.abs().kthvalue(k, dim=self.stats_reduce_dim).values
        return result


class NegativePercentileOrZero(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim', 'q']

    def __init__(self, low_percentile_q, stats_reduce_dim: Optional[int] = None) -> None:
        super(NegativePercentileOrZero, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.q = low_percentile_q
        self.zero = StatelessBuffer(torch.tensor(0.0))

    @brevitas.jit.script_method
    def forward(self, x: Tensor) -> Tensor:
        if self.stats_reduce_dim is None:
            # k is 1-indexed, so round away from zero
            k = int(math.ceil(.01 * self.q * x.numel()))
            result = x.view(-1).kthvalue(k).values
        else:
            # assuming x is two dimensional, get the other dimension
            assert len(x.size()) == 2, "Only 2-dim input is supported."
            other_dim = abs(self.stats_reduce_dim - 1)
            dim_slice = torch.narrow(x, dim=other_dim, start=0, length=1)
            # k is 1-indexed, so round away from zero
            k = int(math.ceil(.01 * self.q * dim_slice.numel()))
            result = x.kthvalue(k, dim=self.stats_reduce_dim).values
        result = torch.where(
            result <= self.zero().to(result.dtype), result, self.zero().to(result.dtype))
        return result


class PercentileInterval(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim', 'low_q', 'high_q']

    def __init__(
            self,
            low_percentile_q,
            high_percentile_q,
            stats_reduce_dim: Optional[int] = None) -> None:
        super(PercentileInterval, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.low_q = low_percentile_q
        self.high_q = high_percentile_q

    @brevitas.jit.script_method
    def forward(self, x: Tensor) -> Tensor:
        if self.stats_reduce_dim is None:
            low_k = int(math.ceil(.01 * self.low_q * x.numel()))
            # k is 1-indexed, so round away from zero
            high_k = int(math.floor(.01 * self.high_q * x.numel() + 0.5))
            low_result = x.view(-1).kthvalue(low_k).values
            high_result = x.view(-1).kthvalue(high_k).values
        else:
            # assuming x is two dimensional, get the other dimension
            assert len(x.size()) == 2, "Only 2-dim input is supported."
            other_dim = abs(self.stats_reduce_dim - 1)
            dim_slice = torch.narrow(x, dim=other_dim, start=0, length=1)
            low_k = int(math.ceil(.01 * self.low_q * dim_slice.numel()))
            # k is 1-indexed, so round away from zero
            high_k = int(math.floor(.01 * self.high_q * dim_slice.numel() + 0.5))
            low_result = x.kthvalue(low_k, dim=self.stats_reduce_dim).values
            high_result = x.kthvalue(high_k, dim=self.stats_reduce_dim).values
        interval = high_result - low_result
        abs_interval = torch.abs(interval)
        return abs_interval


class AbsMax(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(AbsMax, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            return torch.max(torch.abs(x))
        else:
            return torch.max(torch.abs(x), dim=self.stats_reduce_dim)[0]


class AbsMinMax(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(AbsMinMax, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            return torch.abs(torch.max(x) - torch.min(x))
        else:
            max_val = torch.max(x, dim=self.stats_reduce_dim)[0]
            min_val = torch.min(x, dim=self.stats_reduce_dim)[0]
            return torch.abs(max_val - min_val)


class AbsMaxAve(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: int) -> None:
        super(AbsMaxAve, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        return torch.mean(torch.max(torch.abs(x), dim=self.stats_reduce_dim)[0])


class AbsMaxL2(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: int) -> None:
        super(AbsMaxL2, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        per_channel_max = torch.max(torch.abs(x), dim=self.stats_reduce_dim)[0]
        out = torch.norm(per_channel_max, p=2)
        out = out / math.sqrt(per_channel_max.view(-1).shape[0])
        return out


class AbsAve(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(AbsAve, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            return torch.mean(torch.abs(x))
        else:
            return torch.mean(torch.abs(x), dim=self.stats_reduce_dim)


class MeanSigmaStd(brevitas.jit.ScriptModule):

    def __init__(
            self,
            sigma: float,
            stats_reduce_dim: Optional[int] = None,
            std_dev_epsilon: float = DEFAULT_STD_DEV_EPSILON) -> None:
        super(MeanSigmaStd, self).__init__()
        self.impl = _MeanSigmaStdImpl(stats_reduce_dim, std_dev_epsilon)
        self.sigma = StatelessBuffer(torch.tensor(sigma))

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        sigma = self.sigma()
        out = self.impl(x, sigma)
        return out


class _MeanSigmaStdImpl(brevitas.jit.ScriptModule):
    __constants__ = ['stats_reduce_dim', 'output_shape', 'epsilon']

    def __init__(
            self,
            stats_reduce_dim: Optional[int] = None,
            std_dev_epsilon: float = DEFAULT_STD_DEV_EPSILON) -> None:
        super(_MeanSigmaStdImpl, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim
        self.epsilon = std_dev_epsilon

    @brevitas.jit.script_method
    def forward(self, x: Tensor, sigma: Tensor):
        abs_val = torch.abs(x)
        if self.stats_reduce_dim is None:
            mean_val = torch.mean(abs_val)
            std_val = torch.sqrt(torch.var(abs_val) + self.epsilon)
        else:
            mean_val = torch.mean(torch.abs(x), dim=self.stats_reduce_dim)
            std_val = torch.sqrt(torch.var(abs_val, dim=self.stats_reduce_dim) + self.epsilon)
            mean_val = mean_val.view(-1)
            std_val = std_val.view(-1)
        return mean_val + sigma * std_val


class MeanLearnedSigmaStd(brevitas.jit.ScriptModule):

    def __init__(
            self,
            sigma: float,
            stats_output_shape: Tuple[int, ...],
            stats_reduce_dim: Optional[int] = None,
            std_dev_epsilon: float = DEFAULT_STD_DEV_EPSILON) -> None:
        super(MeanLearnedSigmaStd, self).__init__()
        self.impl = _MeanSigmaStdImpl(stats_reduce_dim, std_dev_epsilon)
        if stats_output_shape == SCALAR_SHAPE:
            self.value = Parameter(torch.tensor(sigma))
        else:
            self.value = Parameter(torch.full(stats_output_shape, sigma))

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        sigma = self.sigma.view(self.sigma.shape)  # trick to get a tensor type
        out = self.impl(x, sigma)
        return out

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        value_key = prefix + 'sigma'
        retrocomp_value_key = prefix + 'learned_sigma'
        if retrocomp_value_key in state_dict:  # retrocompatibility
            state_dict[value_key] = state_dict.pop(retrocomp_value_key)
        super(MeanLearnedSigmaStd, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        sigma_key = prefix + 'sigma'
        if config.IGNORE_MISSING_KEYS and sigma_key in missing_keys:
            missing_keys.remove(sigma_key)


class KLMinimizerThreshold(torch.nn.Module):
    """
    Based on:
    https://github.com/apache/incubator-mxnet/blob/master/python/mxnet/contrib/quantization.py
    """

    def __init__(self, signed, bit_width_impl, num_bins=1000 + 1, smoothing_eps=0.0001):
        super(KLMinimizerThreshold, self).__init__()
        self.num_bins = num_bins
        self.smoothing_eps = smoothing_eps
        self.signed = signed
        self.bit_width_impl = bit_width_impl
        self.absmax_impl = AbsMax()

    def smooth_normalize_distribution(self, p, eps):
        is_zeros = (p == 0).float()
        n_zeros = is_zeros.sum()
        n_nonzeros = torch.numel(p) - n_zeros
        if not n_nonzeros:
            return None
        eps1 = eps * n_zeros / n_nonzeros
        hist = p.float()
        hist += eps * is_zeros + (-eps1) * n_nonzeros
        dist = torch.distributions.categorical.Categorical(logits=hist)
        return dist

    def forward(self, x: Tensor):
        absmax = self.absmax_impl(x)
        bit_width = self.bit_width_impl()
        num_quantized_bins = max_int(self.signed, False, bit_width).int()
        thresholds = torch.zeros(self.num_bins // 2 + 1 - num_quantized_bins // 2, device=x.device)
        divergence = torch.zeros_like(thresholds)
        quantized_bins = torch.zeros(num_quantized_bins, device=x.device)
        hist = torch.histc(x, bins=self.num_bins, min=-absmax, max=absmax).int()
        hist_edges = torch.linspace(-absmax, absmax, self.num_bins + 1)
        for i in range(num_quantized_bins // 2, self.num_bins // 2 + 1):
            p_bin_idx_start = self.num_bins // 2 - i
            p_bin_idx_stop = self.num_bins // 2 + i + 1
            thresholds[i - num_quantized_bins // 2] = hist_edges[p_bin_idx_stop]
            sliced_nd_hist = hist[p_bin_idx_start:p_bin_idx_stop]
            p = sliced_nd_hist.clone()
            left_outlier_count = torch.sum(hist[0:p_bin_idx_start])
            p[0] += left_outlier_count
            right_outlier_count = torch.sum(hist[p_bin_idx_stop:])
            p[-1] += right_outlier_count
            is_nonzeros = (sliced_nd_hist != 0).float()
            num_merged_bins = torch.numel(p) // num_quantized_bins
            for j in range(num_quantized_bins):
                start = j * num_merged_bins
                stop = start + num_merged_bins
                quantized_bins[j] = sliced_nd_hist[start:stop].sum()
            quantized_bins[-1] += sliced_nd_hist[num_quantized_bins * num_merged_bins:].sum()
            q = torch.zeros_like(p, dtype=torch.float32, device=x.device)
            for j in range(num_quantized_bins):
                start = j * num_merged_bins
                if j == num_quantized_bins - 1:
                    stop = -1
                else:
                    stop = start + num_merged_bins
                norm = is_nonzeros[start:stop].sum()
                if norm != 0:
                    q[start:stop] = quantized_bins[j] / norm
            q[sliced_nd_hist == 0] = 0.
            p = self.smooth_normalize_distribution(p, self.smoothing_eps)
            q = self.smooth_normalize_distribution(q, self.smoothing_eps)
            if q is None:
                divergence[i - num_quantized_bins // 2] = float('inf')
            else:
                divergence[i - num_quantized_bins // 2] = torch.distributions.kl.kl_divergence(p, q)
        min_divergence_idx = torch.argmin(divergence)
        opt_threshold = thresholds[min_divergence_idx]
        return opt_threshold


class L1Norm(brevitas.jit.ScriptModule):
    """ScriptModule implementation to collect per-channel L1 normalization stats
    for weight normalization-based quantization."""
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(L1Norm, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            # Need to be able to return the max per-channel L1 norm as a scalar
            raise NotImplementedError("L1 normalization is not supported per-tensor yet.")
        else:
            return x.norm(p=1, dim=self.stats_reduce_dim, keepdim=True)


class L2Norm(brevitas.jit.ScriptModule):
    """ScriptModule implementation to collect per-channel L2 normalization stats
    for weight normalization-based quantization."""
    __constants__ = ['stats_reduce_dim']

    def __init__(self, stats_reduce_dim: Optional[int] = None) -> None:
        super(L2Norm, self).__init__()
        self.stats_reduce_dim = stats_reduce_dim

    @brevitas.jit.script_method
    def forward(self, x: Tensor):
        if self.stats_reduce_dim is None:
            # Need to be able to return the max per-channel L2 norm as a scalar
            raise NotImplementedError("L2 normalization is not supported per-tensor yet.")
        else:
            return x.norm(p=2, dim=self.stats_reduce_dim, keepdim=True)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from collections import namedtuple
from copy import deepcopy
from functools import partial
import operator
from typing import Callable, Dict, List, Optional, Set, Tuple, Union

import torch
import torch.nn as nn

from brevitas.fx import GraphModule
from brevitas.fx import Node
from brevitas.graph.utils import get_module

from .base import GraphTransform

__all__ = ['EqualizeGraph']

EPSILON = 1e-9

Region = namedtuple('Region', ['srcs', 'sinks'])

_supported_layers = (
    nn.ConvTranspose1d,
    nn.ConvTranspose2d,
    nn.ConvTranspose3d,
    nn.MultiheadAttention,
    nn.Conv1d,
    nn.Conv2d,
    nn.Conv3d,
    nn.Linear,
    nn.LayerNorm,
    nn.BatchNorm1d,
    nn.BatchNorm2d,
    nn.BatchNorm3d)

_scale_invariant_layers = (
    nn.Dropout,
    nn.Dropout2d,
    nn.Dropout3d,
    nn.ReLU,
    nn.MaxPool1d,
    nn.MaxPool2d,
    nn.MaxPool3d,
    nn.AvgPool1d,
    nn.AvgPool2d,
    nn.AvgPool3d,
    nn.AdaptiveAvgPool1d,
    nn.AdaptiveAvgPool2d,
    nn.AdaptiveAvgPool3d)

_scale_invariant_op = (torch.mul, operator.mul, operator.imul, operator.__mul__, operator.__imul__)

_select_op = (operator.getitem, operator.__getitem__)

_residual_methods = ('add', 'add_')

_residual_fns = (torch.add, operator.add, operator.iadd, operator.__add__, operator.__iadd__)

_batch_norm = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)

WeightBiasTuple = namedtuple('WeightBiasTuple', ['weight', 'bias'], defaults=[None])


def _select_scale_computation_fn(
        scale_computation_type: str) -> Callable[[torch.Tensor], torch.Tensor]:
    if scale_computation_type == 'maxabs':
        return _channel_maxabs
    elif scale_computation_type == 'range':
        return _channel_range
    else:
        raise RuntimeError(f"Scale computation type {scale_computation_type} not supported")


def _channel_range(inp: torch.Tensor) -> torch.Tensor:
    mins, _ = inp.min(dim=1)
    maxs, _ = inp.max(dim=1)
    out = maxs - mins
    # correct corner case where where all weights along a channel have the same value
    # e.g. when a mean/nn.AvgPool/nn.AdaptiveAvgPool is converted to a depth-wise conv
    out = torch.where(out == 0., torch.mean(inp, dim=1), out)

    # convert to positive range, in case any of the values are negative,
    # highly likely in cases when there is only one value per channel, such as in Batch Norm
    out = torch.abs(out)
    return out


def _channel_maxabs(inp: torch.Tensor) -> torch.Tensor:
    out = torch.max(torch.abs(inp), dim=1)[0]
    return out


def _get_size(axes: Dict[nn.Module, int]) -> int:
    m0, axis0 = list(axes.items())[0]
    size = m0.weight.size(axis0)
    for m, axis in axes.items():
        if m.weight.size(axis) != size:
            return None
    return size


def _get_input_axis(module: nn.Module) -> Optional[int]:
    """
    Given a sink module, determine the axis associated to the input channels.
    Return None if not supported.
    """
    if isinstance(module, (nn.Linear, nn.MultiheadAttention)):
        return 1
    elif isinstance(module, _batch_norm):
        return 0
    elif isinstance(module, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        if module.groups == 1:
            return 1
        elif module.groups == module.out_channels:
            return 0
    elif isinstance(module, (nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)):
        if module.groups == 1:
            return 0
        elif module.groups == module.out_channels:
            return 1
    elif isinstance(module, nn.LayerNorm):
        # We assume normalization happens only along the channel dimension
        if len(module.weight.shape) == 1:
            return 0
        else:
            return None
    else:
        return None


def _get_output_axis(module: nn.Module) -> Optional[int]:
    """
    Given a source module, determine the axis associated to the output channels.
    Return None if not supported.
    """
    if isinstance(module,
                  (nn.Linear,
                   nn.Conv1d,
                   nn.Conv2d,
                   nn.Conv3d,
                   nn.MultiheadAttention,
                   nn.BatchNorm1d,
                   nn.BatchNorm2d,
                   nn.BatchNorm3d)):
        return 0
    elif isinstance(module, (nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)):
        return 1
    elif isinstance(module, nn.LayerNorm):
        # We assume normalization happens only along the channel dimension
        if len(module.weight.shape) == 1:
            return 0
        else:
            return None
    else:
        return None


def _combine_weights_bias(
        weight: torch.Tensor, bias_shrinkage: Union[float, str],
        bias: Optional[torch.Tensor]) -> torch.Tensor:
    """Combine weights and bias before graph equalizattion
    This method merges the weight and bias of the sources, so that the resulting equalizer scale factor
    is influenced also by the magnitude of the bias, mitigated by a shrink factor.
    This technique avoids that, after the equalization procedure, the bias values become too big,
    negatively impacting the quantization accuracy.
    The bias shrinkage factor regulates how much the bias magnitude will affect the subsequence
    calculation of the scale.
    """
    if bias is None:
        return weight.data

    bias = bias.data
    weight = deepcopy(weight).view(weight.shape[0], -1)
    bias = deepcopy(bias).view(-1, 1)

    weight = torch.where(torch.abs(weight) < EPSILON, torch.tensor(EPSILON).type_as(weight), weight)
    factor = torch.abs(bias) / torch.abs(weight)

    # From https://github.com/Xilinx/Vitis-AI/blob/master/src/vai_quantizer/vai_q_pytorch/nndct_shared/optimization/commander.py#L450
    if bias_shrinkage == 'vaiq':
        if torch.abs(bias).max() < 10 and (torch.abs(bias).max() / torch.abs(weight).max()) < 20:
            if torch.median(factor) > 100 or torch.mean(factor) > 1000:
                shrink_factor = 5.
            else:
                shrink_factor = 2.
        else:
            if torch.median(factor) > 30 or torch.mean(factor) > 500:
                shrink_factor = 20.
            elif torch.median(factor) > 15 or torch.mean(factor) > 250:
                shrink_factor = 10.
            else:
                shrink_factor = 5.
    elif isinstance(bias_shrinkage, (int, float)):
        shrink_factor = bias_shrinkage
    else:
        raise RuntimeError(f"{bias_shrinkage} not supported.")
    weight_bias = torch.cat([weight, bias / shrink_factor], 1)
    return weight_bias


def _cross_layer_equalization(
        srcs: List[nn.Module],
        sinks: List[nn.Module],
        merge_bias: bool,
        bias_shrinkage: Union[float, str],
        scale_computation_type: str) -> torch.Tensor:
    """
    Given two adjacent tensors', the weights are scaled such that
    the ranges of the first tensors' output channel are equal to the
    ranges of the second tensors' input channel
    """
    # Determine device and type of tensors
    device = next(srcs[0].parameters()).device
    dtype = next(srcs[0].parameters()).dtype

    src_axes = {}
    sink_axes = {}

    for i, module in enumerate(srcs):
        # If module is not supported, do not perform graph equalization
        if not isinstance(module, _supported_layers):
            return torch.tensor(1., dtype=dtype, device=device)
        if isinstance(module, nn.MultiheadAttention):
            srcs[i] = module.out_proj
        src_axes[srcs[i]] = _get_output_axis(module)

    for i, module in enumerate(sinks):
        # If module is not supported, do not perform graph equalization
        if not isinstance(module, _supported_layers):
            return torch.tensor(1., dtype=dtype, device=device)
        # For MultiheadAttention, we support only self-attetion
        if isinstance(module, nn.MultiheadAttention) and hasattr(sinks[i], 'in_proj_weight'):
            # For sinks, we only need to modify the weight but not the bias
            sinks[i] = WeightBiasTuple(module.in_proj_weight)
        elif isinstance(module, nn.MultiheadAttention) and not hasattr(sinks[i], 'in_proj_weight'):
            return torch.tensor(1., dtype=dtype, device=device)
        sink_axes[sinks[i]] = _get_input_axis(module)

    # Check if any of the axis is None, which means that the module is not supported.
    # In that case, do not perform graph equalization
    if None in [*src_axes.values(), *sink_axes.values()]:
        return torch.tensor(1., dtype=dtype, device=device)

    src_size = _get_size(src_axes)
    sink_size = _get_size(sink_axes)

    # Check if any of the src_size or sink_size is None, which means that the some of the
    # sources or sinks do not have the same size as the others.
    # Similarly, exit if source and sink have different different sizes
    if None in [src_size, sink_size] or src_size != sink_size:
        return torch.tensor(1., dtype=dtype, device=device)
    transpose = lambda module, axis: module.weight if axis == 0 else module.weight.transpose(0, 1)
    scale_fn = _select_scale_computation_fn(scale_computation_type)
    if merge_bias:
        src_weights = [
            _combine_weights_bias(transpose(m, axis), bias_shrinkage, m.bias) for m,
            axis in src_axes.items()]
    else:
        src_weights = [transpose(m, axis) for m, axis in src_axes.items()]
    sink_weights = [transpose(m, axis) for m, axis in sink_axes.items()]
    srcs_range = scale_fn(torch.cat([w.reshape(w.size(0), -1) for w in src_weights], 1))
    sinks_range = scale_fn(torch.cat([w.reshape(w.size(0), -1) for w in sink_weights], 1))
    sinks_range += EPSILON

    scaling_factors = torch.sqrt(srcs_range / sinks_range)
    inverse_scaling_factors = torch.reciprocal(scaling_factors)

    for module, axis in src_axes.items():
        if hasattr(module, 'bias') and module.bias is not None:
            module.bias.data = module.bias.data * inverse_scaling_factors.view_as(module.bias)
        src_broadcast_size = [1] * module.weight.ndim
        src_broadcast_size[axis] = module.weight.size(axis)
        module.weight.data = module.weight.data * torch.reshape(
            inverse_scaling_factors, src_broadcast_size)
    for module, axis in sink_axes.items():
        src_broadcast_size = [1] * module.weight.ndim
        src_broadcast_size[axis] = module.weight.size(axis)
        if isinstance(module, _batch_norm):
            # We re-compute the bias as function of running_mean and running_var to adjust the
            # additive factor for equalization.
            additive_factor = module.running_mean.data * module.weight.data / torch.sqrt(
                module.running_var.data + module.eps)
            module.bias.data = module.bias.data + additive_factor * (scaling_factors - 1)
        module.weight.data = module.weight.data * torch.reshape(scaling_factors, src_broadcast_size)

    return scaling_factors


def _equalize(
        model: GraphModule,
        regions: Set[Tuple[str]],
        iterations: int,
        threshold: float,
        merge_bias: bool,
        bias_shrinkage: Union[str, float],
        scale_computation_type: str) -> GraphModule:
    """
    Generalized version of section 4.1 of https://arxiv.org/pdf/1906.04721.pdf
    """
    name_to_module: Dict[str, nn.Module] = {}
    name_set = {name for region in regions for module_set in region for name in module_set}

    for name, module in model.named_modules():
        if name in name_set:
            name_to_module[name] = module
    for i in range(iterations):
        scale_factor_max = None
        for region in regions:
            scale_factors_region = _cross_layer_equalization(
                [name_to_module[n] for n in region.srcs], [name_to_module[n] for n in region.sinks],
                merge_bias,
                bias_shrinkage,
                scale_computation_type)
            scale_factor_region_max = torch.max(torch.abs(1 - scale_factors_region))
            if scale_factor_max is not None:
                scale_factor_max = torch.max(scale_factor_max, scale_factor_region_max)
            else:
                scale_factor_max = scale_factor_region_max
        if threshold is not None and scale_factor_max < threshold:
            break
    return model


def _is_supported_module(graph_model: GraphModule, node: Node) -> bool:
    if node.op == 'call_module':
        module = get_module(graph_model, node.target)
        if isinstance(module, _supported_layers):
            # We support only self-attention
            if isinstance(module, nn.MultiheadAttention):
                return all([node.all_input_nodes[0].name == n.name for n in node.all_input_nodes])
            return True
    return False


def _is_scale_invariant_module(graph_model: GraphModule, node: Node) -> bool:
    return node.op == 'call_module' and isinstance(
        get_module(graph_model, node.target), _scale_invariant_layers)


def _is_scale_invariant_function(node: Node) -> bool:
    return node.op == 'call_function' and node.target in _scale_invariant_op + _select_op


def _is_reshaping_op(node: Node) -> bool:
    return (
        node.op == 'call_function' and node.target in [torch.flatten, torch.reshape] or
        node.op == 'call_method' and node.target in ['view', 'reshape', 'flatten'])


def walk_region(
        graph_model: GraphModule,
        starting_node: Node,
        history: Set[Node],
        srcs: Set[str],
        sinks: Set[str],
        walk_forward: bool):
    node_list = starting_node.users if walk_forward else starting_node.all_input_nodes
    for node in node_list:
        # we keep a history of how the graph has been walked already, invariant to the direction,
        # to avoid getting stuck in a loop
        path = (starting_node, node) if walk_forward else (node, starting_node)
        if path not in history:
            history.add(path)
        else:
            continue
        if _is_supported_module(graph_model, node):
            if walk_forward:
                module = get_module(graph_model, node.target)
                # It is not possible to equalize through LayerNorm as sink
                if not isinstance(module, nn.LayerNorm):
                    sinks.add(node.target)
            else:
                srcs.add(node.target)
                walk_region(graph_model, node, history, srcs, sinks, walk_forward=True)
        elif _is_scale_invariant_module(graph_model, node) or _is_scale_invariant_function(node):
            if walk_forward:
                walk_region(graph_model, node, history, srcs, sinks, walk_forward=True)
            else:
                walk_region(graph_model, node, history, srcs, sinks, walk_forward=True)
                walk_region(graph_model, node, history, srcs, sinks, walk_forward=False)
        elif (node.op == 'call_method' and node.target in _residual_methods or
              node.op == 'call_function' and node.target in _residual_fns):
            walk_region(graph_model, node, history, srcs, sinks, walk_forward=True)
            walk_region(graph_model, node, history, srcs, sinks, walk_forward=False)
        elif _is_reshaping_op(node):
            walk_region(graph_model, node, history, srcs, sinks, walk_forward=walk_forward)
        else:
            continue


def _extract_regions(graph_model: GraphModule) -> Set[Tuple[str]]:
    regions = set()
    for node in graph_model.graph.nodes:
        if _is_supported_module(graph_model, node):
            srcs, sinks = {node.target}, set()
            walk_region(graph_model, node, set(), srcs, sinks, walk_forward=True)
            if sinks:
                # each region should appear only once, so to make it hashable
                # we convert srcs and sinks to ordered lists first, and then to tuples
                regions.add(Region(tuple(sorted(srcs)), tuple(sorted(sinks))))
    return regions


class EqualizeGraph(GraphTransform):

    def __init__(
            self,
            iterations: int = 10,
            threshold: float = 0.05,
            return_regions: bool = False,
            merge_bias: bool = True,
            bias_shrinkage: Union[float, str] = 'vaiq',
            scale_computation_type: str = 'maxabs') -> None:
        super(EqualizeGraph, self).__init__()
        self.iterations = iterations
        self.return_regions = return_regions
        self.merge_bias = merge_bias
        self.bias_shrinkage = bias_shrinkage
        self.threshold = threshold
        self.scale_computation_type = scale_computation_type

    def apply(self,
              graph_model: GraphModule) -> Union[Tuple[GraphModule, Set[Tuple[str]]], GraphModule]:
        regions = _extract_regions(graph_model)
        if len(regions) > 0:
            graph_model = _equalize(
                graph_model,
                regions,
                self.iterations,
                self.threshold,
                self.merge_bias,
                self.bias_shrinkage,
                self.scale_computation_type)
        if self.return_regions:
            return graph_model, regions
        else:
            return graph_model


class AbsorbBiasByBatchNorm(GraphTransform):

    def __init__(self):
        super(AbsorbBiasByBatchNorm, self).__init__()
        self.inp_shape_map = {}
        self.collect_inp_shape_hooks = []

    def add_to_bias(self, module, tensor):
        if module.bias is not None:
            module.bias.data += tensor.view_as(module.bias)
        else:
            module.bias = nn.Parameter(tensor)

    def absorb_biases(self, groups):
        for layer, bn, (next_layer_name, next_layer) in groups:
            cfactor = bn.running_mean - 3 * torch.sqrt(bn.running_var)
            zeroes = torch.zeros_like(cfactor).to(cfactor.device)
            cfactor = torch.where(cfactor > 0., cfactor, zeroes)
            if (cfactor > 0).any():
                self.add_to_bias(layer, -cfactor)
                broadcast_shape = [1] * next_layer.weight.ndim
                broadcast_shape[1] = cfactor.numel()
                cfactor = cfactor.view(broadcast_shape)
                cfactor = cfactor.expand(self.inp_shape_map[next_layer_name])
                next_layer_cfactor = next_layer(cfactor).transpose(0, 1)
                next_layer_cfactor = next_layer_cfactor.view(next_layer_cfactor.shape[0], -1)
                next_layer_cfactor = torch.mean(next_layer_cfactor, dim=1)
                self.add_to_bias(next_layer, next_layer_cfactor)
        self.inp_shape_map = {}

    def extract_groups(self, graph_model: GraphModule):
        groups = []
        for node in graph_model.graph.nodes:
            if (_is_supported_module(graph_model, node) and node.next.op == 'call_module' and
                    isinstance(get_module(graph_model, node.next.target), _batch_norm)):
                node_next = node.next.next
                while _is_scale_invariant_module(graph_model,
                                                 node_next) or _is_reshaping_op(node_next):
                    node_next = node_next.next
                if _is_supported_module(graph_model, node_next):
                    group = (
                        get_module(graph_model, node.target),
                        get_module(graph_model, node.next.target),
                        (node_next.target, get_module(graph_model, node_next.target)))
                    groups.append(group)
        return groups

    def collect_inp_shape_hook(self, module, inp, name):
        if name in self.inp_shape_map.keys():
            raise RuntimeError("Module called multiple times, not supported.")
        if isinstance(inp, tuple):
            inp = inp[0]
        self.inp_shape_map[name] = [1] + list(inp.shape[1:])

    def collect_inp_shapes(self, model, inp):
        for name, module in model.named_modules():
            if isinstance(module, (_supported_layers)):
                hook_fn = partial(self.collect_inp_shape_hook, name=name)
                hook = module.register_forward_pre_hook(hook_fn)
                self.collect_inp_shape_hooks.append(hook)
        model(inp)
        for hook in self.collect_inp_shape_hooks:
            hook.remove()
        self.collect_inp_shape_hooks = []

    def apply(self, graph_model: GraphModule, inp):
        self.collect_inp_shapes(graph_model, inp)
        groups = self.extract_groups(graph_model)
        self.absorb_biases(groups)
        return graph_model



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

# retrocompatibility
# the assert prevents the removal of the unused import
from brevitas.inject.enum import BitWidthImplType

from .const import BitWidthConst
from .const import BitWidthStatefulConst
from .const import MsbClampBitWidth
from .parameter import BitWidthParameter
from .parameter import RemoveBitwidthParameter

assert BitWidthImplType



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.quant import QuantType
from brevitas.core.restrict_val import RestrictValueType
from brevitas.core.scaling import ScalingImplType
from brevitas.core.stats import StatsOp
import brevitas.nn as quant_nn

global ACT_MIN_VAL, ACT_MAX_VAL
brevitas_activations = {
    "hardtanh": quant_nn.QuantHardTanh,
    "relu": quant_nn.QuantReLU,}

QUANT_TYPE = QuantType.INT
QUANT_TYPE_BIAS = QuantType.FP

SCALING_MIN_VAL = 2e-16
ACT_SCALING_IMPL_TYPE = ScalingImplType.PARAMETER
ACT_SCALING_PER_CHANNEL = False
ACT_RESTRICT_SCALING_TYPE = RestrictValueType.LOG_FP

WEIGHT_SCALING_IMPL_TYPE = ScalingImplType.STATS
WEIGHT_SCALING_STATS_OP = StatsOp.MAX
WEIGHT_NARROW_RANGE = True
BIAS_CONFIGS = False


def make_quantization_input(bit_width, absolute_act_val, scaling_per_channel):
    return quant_nn.QuantHardTanh(
        bit_width=bit_width,
        scaling_per_output_channel=scaling_per_channel,
        quant_type=QUANT_TYPE,
        scaling_impl_type=ACT_SCALING_IMPL_TYPE,
        scaling_min_val=SCALING_MIN_VAL,
        restrict_scaling_type=ACT_RESTRICT_SCALING_TYPE,
        max_val=absolute_act_val,
        min_val=-absolute_act_val,
        return_quant_tensor=False)


def make_norm_scale(bit_width, absolute_act_val, scaling_per_channel):
    return quant_nn.QuantHardTanh(
        bit_width=bit_width,
        scaling_per_output_channel=scaling_per_channel,
        quant_type=QUANT_TYPE,
        scaling_impl_type=ACT_SCALING_IMPL_TYPE,
        scaling_min_val=SCALING_MIN_VAL,
        restrict_scaling_type=ACT_RESTRICT_SCALING_TYPE,
        max_val=absolute_act_val,
        min_val=-absolute_act_val,
        scaling_stats_permute_dims=(1, 0, 2),
        return_quant_tensor=True)


def make_jasper_activation(activation, channels, bit_width, absolute_act_val, scaling_per_channel):
    brevitas_activation = brevitas_activations[activation]
    return brevitas_activation(
        bit_width=bit_width,
        scaling_per_output_channel=scaling_per_channel,
        quant_type=QUANT_TYPE,
        scaling_impl_type=ACT_SCALING_IMPL_TYPE,
        scaling_min_val=SCALING_MIN_VAL,
        restrict_scaling_type=ACT_RESTRICT_SCALING_TYPE,
        max_val=absolute_act_val,
        per_channel_broadcastable_shape=(1, channels, 1),
        scaling_stats_permute_dims=(1, 0, 2),
        return_quant_tensor=False)


def make_quantconv1d(
        feat_in,
        classes,
        kernel_size,
        bit_width,
        scaling_per_channel,
        bias,
        stride=1,
        padding=0,
        dilation=1,
        groups=1):

    return quant_nn.QuantConv1d(
        in_channels=feat_in,
        out_channels=classes,
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=groups,
        bias=bias,
        weight_bit_width=bit_width,
        weight_scaling_stats_op=WEIGHT_SCALING_STATS_OP,
        weight_scaling_per_output_channel=scaling_per_channel,
        weight_quant_type=QUANT_TYPE,
        weight_narrow_range=WEIGHT_NARROW_RANGE,
        weight_scaling_impl_type=WEIGHT_SCALING_IMPL_TYPE,
        weight_scaling_min_val=SCALING_MIN_VAL,
        bias_bit_width=bit_width,
        bias_quant_type=QUANT_TYPE_BIAS,
        bias_narrow_range=BIAS_CONFIGS,
        compute_output_scale=BIAS_CONFIGS,
        compute_output_bit_width=BIAS_CONFIGS,
        return_quant_tensor=False)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from abc import abstractmethod
from typing import Optional, Type, Union
from warnings import warn

from torch.nn import Module

from brevitas.inject import ExtendedInjector
from brevitas.inject import Injector
from brevitas.proxy.runtime_quant import ActQuantProxyFromInjector
from brevitas.proxy.runtime_quant import ActQuantProxyProtocol
from brevitas.quant import NoneActQuant

from .base import QuantProxyMixin

ActQuantType = Union[ActQuantProxyProtocol, Type[Injector], Type[ExtendedInjector]]


class QuantInputMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(self, input_quant: Optional[ActQuantType], **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=input_quant,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            proxy_prefix='input_',
            kwargs_prefix='input_',
            input_act_impl=None,
            input_passthrough_act=True,
            **kwargs)

    @property
    def is_input_quant_enabled(self):
        return self.input_quant.is_quant_enabled

    @property
    def is_quant_input_narrow_range(self):  # TODO make abstract once narrow range can be cached
        return self.input_quant.is_narrow_range

    @property
    @abstractmethod
    def is_quant_input_signed(self):
        pass

    @abstractmethod
    def quant_input_scale(self):
        pass

    @abstractmethod
    def quant_input_zero_point(self):
        pass

    @abstractmethod
    def quant_input_bit_width(self):
        pass


class QuantOutputMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(self, output_quant: Optional[ActQuantType], **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=output_quant,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            proxy_prefix='output_',
            kwargs_prefix='output_',
            output_act_impl=None,
            output_passthrough_act=True,
            **kwargs)

    @property
    def is_output_quant_enabled(self):
        return self.output_quant.is_quant_enabled

    @property
    def is_quant_output_narrow_range(self):  # TODO make abstract once narrow range can be cached
        return self.output_quant.is_narrow_range

    @property
    @abstractmethod
    def is_quant_output_signed(self):
        pass

    @abstractmethod
    def quant_output_scale(self):
        pass

    @abstractmethod
    def quant_output_zero_point(self):
        pass

    @abstractmethod
    def quant_output_bit_width(self):
        pass


class QuantNonLinearActMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            act_impl: Optional[Type[Module]],
            passthrough_act: bool,
            act_quant: Optional[ActQuantType],
            act_proxy_prefix='act_',
            act_kwargs_prefix='',
            **kwargs):
        prefixed_kwargs = {
            act_kwargs_prefix + 'act_impl': act_impl,
            act_kwargs_prefix + 'passthrough_act': passthrough_act}
        QuantProxyMixin.__init__(
            self,
            quant=act_quant,
            proxy_prefix=act_proxy_prefix,
            kwargs_prefix=act_kwargs_prefix,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            **prefixed_kwargs,
            **kwargs)

    @property
    def is_act_quant_enabled(self):
        return self.act_quant.is_quant_enabled

    @property
    def is_quant_act_narrow_range(self):  # TODO make abstract once narrow range can be cached
        return self.act_quant.is_narrow_range

    @property
    @abstractmethod
    def is_quant_act_signed(self):
        pass

    @abstractmethod
    def quant_act_scale(self):
        pass

    @abstractmethod
    def quant_act_zero_point(self):
        pass

    @abstractmethod
    def quant_act_bit_width(self):
        pass



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from typing import Union
import warnings

import torch
from torch import Tensor

from brevitas.nn import QuantConv1d
from brevitas.nn import QuantConv2d
from brevitas.nn import QuantLinear
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL

from .base import PytorchQuantLayerHandler


class PytorchQuantWBIOLHandler(PytorchQuantLayerHandler):

    def __init__(self):
        super().__init__()
        self.qf_impl = None
        self.input_quant_impl = None
        self.weight_quant_impl = None
        self.weight_quant_args = None
        self.input_quant_kwargs = None
        self.weight_quant_kwargs = None
        self.output_quant_kwargs = None
        self.qf_kwargs = None
        self.return_quant_tensor = None

    @classmethod
    def validate(cls, module: QuantWBIOL):
        assert module.is_weight_quant_enabled, 'Weight quantization required'
        assert module.is_output_quant_enabled, 'Output quantization required'

    @classmethod
    def prepare_bias(cls, module: QuantWBIOL):
        if module.bias is not None and not module.is_bias_quant_enabled:
            bias = module.bias.detach()
        elif module.bias is not None and module.is_bias_quant_enabled:
            # export the dequantized value
            bias = module.quant_bias().value
        else:
            bias = module.bias
        return bias

    @classmethod
    def prepare_weight_quant(cls, module: QuantWBIOL):
        cls.validate_bit_width(module.quant_weight_bit_width(), 7, le_then=True)
        cls.validate_8b_bit_width(module.quant_input_bit_width(), le_then=False)
        cls.validate_8b_bit_width(module.quant_output_bit_width(), le_then=False)
        scale = module.quant_weight_scale()
        zero_point = cls.quant_weight_zero_point(module)
        signed = module.is_quant_weight_signed
        weight = module.weight.detach()
        quant_impl, quant_kwargs = cls.gen_quant_impl_kwargs(scale, zero_point, signed)
        return quant_impl, (weight,), quant_kwargs

    def prepare_for_export(self, module: QuantWBIOL):
        self.validate(module)
        if module.is_input_quant_enabled:
            self.input_quant_impl, self.input_quant_kwargs = self.prepare_input_quant(module)
        weight_quant_pack = self.prepare_weight_quant(module)
        self.weight_quant_impl, self.weight_quant_args, self.weight_quant_kwargs = weight_quant_pack
        self.qf_impl, self.qf_kwargs = self.prepare_qf(module)
        _, self.output_quant_kwargs = self.prepare_output_quant(module)
        self.return_quant_tensor = module.return_quant_tensor

    def q_weight(self):
        q_weight = self.weight_quant_impl(*self.weight_quant_args, **self.weight_quant_kwargs)
        return q_weight

    def forward(self, q_inp: Tensor):
        if self.input_quant_impl is not None:
            # If the input is quantized from a previous layer,
            # we have to dequant and requant
            if q_inp.is_quantized:
                q_inp.dequantize()
            q_inp = self.input_quant_impl(q_inp, **self.input_quant_kwargs)
        assert q_inp.is_quantized, 'Input needs to be quantized'
        q_out = self.qf_impl(q_inp, self.q_weight(), **self.qf_kwargs, **self.output_quant_kwargs)
        if not self.return_quant_tensor:
            q_out = q_out.dequantize()
        return q_out


class PytorchQuantConvNdHandler(PytorchQuantWBIOLHandler, ABC):

    @classmethod
    def explicit_output_dtype(cls):
        return True

    @classmethod
    def prepare_qf_kwargs(cls, module: Union[QuantConv1d, QuantConv2d]):
        return {
            'bias': cls.prepare_bias(module),
            'stride': module.stride,
            'padding': module.padding,
            'dilation': module.dilation,
            'groups': module.groups,
            'padding_mode': module.padding_mode}


class PytorchQuantConv1dHandler(PytorchQuantConvNdHandler):
    handled_layer = QuantConv1d

    @classmethod
    def prepare_qf(cls, module: QuantConv1d):
        return torch.nn.quantized.functional.conv1d, cls.prepare_qf_kwargs(module)


class PytorchQuantConv2dHandler(PytorchQuantConvNdHandler):
    handled_layer = QuantConv2d

    @classmethod
    def prepare_qf(cls, module: QuantConv2d):
        return torch.nn.quantized.functional.conv2d, cls.prepare_qf_kwargs(module)


class PytorchQuantLinearHandler(PytorchQuantWBIOLHandler):
    handled_layer = QuantLinear

    @classmethod
    def explicit_output_dtype(cls):
        return False

    @classmethod
    def prepare_qf(cls, module: QuantLinear):
        return torch.nn.quantized.functional.linear, {'bias': cls.prepare_bias(module)}



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Union

from torch import Tensor
from torch.nn import MaxPool1d
from torch.nn import MaxPool2d

from brevitas.quant_tensor import QuantTensor

from .mixin.base import QuantLayerMixin


class QuantMaxPool1d(QuantLayerMixin, MaxPool1d):

    def __init__(
            self,
            kernel_size,
            stride=None,
            padding=0,
            dilation=1,
            return_indices=False,
            ceil_mode=False,
            return_quant_tensor: bool = True):
        MaxPool1d.__init__(
            self,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            return_indices=return_indices,
            ceil_mode=ceil_mode)
        QuantLayerMixin.__init__(self, return_quant_tensor=return_quant_tensor)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return False

    def forward(self, input: Union[Tensor, QuantTensor]):
        x = self.unpack_input(input)
        if self.export_mode:
            return self.export_handler(x.value)
        x = x.set(value=super().forward(x.value))
        return self.pack_output(x)


class QuantMaxPool2d(QuantLayerMixin, MaxPool2d):

    def __init__(
            self,
            kernel_size,
            stride=None,
            padding=0,
            dilation=1,
            return_indices=False,
            ceil_mode=False,
            return_quant_tensor: bool = True):
        MaxPool2d.__init__(
            self,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            return_indices=return_indices,
            ceil_mode=ceil_mode)
        QuantLayerMixin.__init__(self, return_quant_tensor=return_quant_tensor)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return False

    def forward(self, input: Union[Tensor, QuantTensor]):
        x = self.unpack_input(input)
        if self.export_mode:
            out = self.export_handler(x.value)
            self._set_global_is_quant_layer(False)
            return out
        x = x.set(value=super().forward(x.value))
        return self.pack_output(x)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from contextlib import contextmanager
from enum import Enum


class AutoName(str, Enum):

    def _generate_next_value_(name, start, count, last_values):
        return name

    def __str__(self):
        return self.value

    def __eq__(self, other):
        return str(self).lower() == str(other).lower()


@contextmanager
def patch(module, attr, new_value):
    """
    Context manager for monkey patching.
    Each patch can be used only once.
    with patch(torch, 'add', custom_add):
        ...
    """
    old_value = getattr(module, attr)
    setattr(module, attr, new_value)
    try:
        yield getattr(module, attr)
    finally:
        setattr(module, attr, old_value)


def islambda(v):
    LAMBDA = lambda: 0
    return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Tuple

import torch
from torch import Tensor

from brevitas.nn import QuantHardTanh
from brevitas.nn import QuantIdentity
from brevitas.nn import QuantReLU

from ..function.act import QuantHardTanhFn
from ..function.act import QuantReLUFn
from ..utils import finn_datatype
from .base import FINNQuantInputHandler


class FINNQuantReLUHandler(FINNQuantInputHandler):
    handled_layer = QuantReLU

    @staticmethod
    def quant_type(module: QuantReLU):
        bit_width = module.quant_act_bit_width()
        signed = module.is_quant_act_signed
        return finn_datatype(bit_width, signed)

    @staticmethod
    def thresholds(module: QuantReLU, extend_tensor_to_channels=True):
        num_distinct_values = 2 ** int(module.quant_act_bit_width().item())
        num_thresholds = num_distinct_values - 1
        flat_scale = module.quant_act_scale().view(-1)
        num_scale_channels = flat_scale.shape[0]
        step = torch.abs(flat_scale)
        min_threshold = step / 2
        thresholds = torch.empty(num_scale_channels, num_thresholds)
        for c in range(num_scale_channels):
            for t in range(num_thresholds):
                thresholds[c][t] = min_threshold[c] + step[c] * t
        if extend_tensor_to_channels:
            output_channels = module._cached_inp.shape[1]
            final_shape = (output_channels, num_thresholds)
            if thresholds.shape != final_shape:
                thresholds = thresholds.expand(final_shape)
        return thresholds

    @staticmethod
    def quant_act_scale(module: QuantReLU):
        quant_act_scale = module.quant_act_scale().type(torch.FloatTensor).detach()
        return quant_act_scale

    def prepare_for_export(self, module: QuantReLU):
        self.symbolic_kwargs = {
            'qnt_type': self.quant_type(module),
            'thres': self.thresholds(module),
            'bias': None,
            'scale': self.quant_act_scale(module)}

    def symbolic_execution(self, inp: Tensor):
        ret = QuantReLUFn.apply(inp, *self.symbolic_kwargs.values())
        return ret


class FINNQuantHardTanhHandler(FINNQuantInputHandler):
    handled_layer = QuantHardTanh

    @staticmethod
    def quant_type(module: QuantHardTanh):
        return finn_datatype(module.quant_act_bit_width(), module.is_quant_act_signed)

    @staticmethod
    def quant_act_bias(module: QuantHardTanh):
        bit_width = int(module.quant_act_bit_width().item())
        if bit_width == 1:
            return torch.tensor(-0.5).type(torch.FloatTensor)
        else:
            if module.is_quant_act_narrow_range:
                min_non_scaled_val = -(2 ** (bit_width - 1) - 1)
            else:
                min_non_scaled_val = -2 ** (bit_width - 1)
            return torch.tensor(min_non_scaled_val).type(torch.FloatTensor)

    @staticmethod
    def thresholds(module: QuantHardTanh, extend_tensor_to_channels=True):
        bit_width = int(module.quant_act_bit_width().item())
        if bit_width != 1:
            if module.is_quant_act_narrow_range:
                # assuming narrow range, symmetric quantization around zero
                # when using narrow range, we represent one element less
                num_distinct_values = 2 ** bit_width - 1
            else:
                num_distinct_values = 2 ** bit_width
            num_thresholds = num_distinct_values - 1
            flat_scale = module.quant_act_scale().view(-1)
            num_scale_channels = flat_scale.shape[0]
            step = torch.abs(flat_scale)
            half_step = step / 2.0
            thresholds = torch.empty(num_scale_channels, num_thresholds)
            # compute the value of the smallest threshold, we'll neg-bias all
            # generated thresholds by this much
            min_threshold = -half_step - step * ((num_thresholds // 2) - 1)
            if not module.is_quant_act_narrow_range:
                min_threshold -= step
            for c in range(num_scale_channels):
                for t in range(num_thresholds):
                    thresholds[c][t] = min_threshold[c] + step[c] * t
            if extend_tensor_to_channels:
                output_channels = module._cached_inp.shape[1]
                final_shape = (output_channels, num_thresholds)
                if thresholds.shape != final_shape:
                    thresholds = thresholds.expand(final_shape)
            return thresholds
        else:
            thresholds = torch.empty([1, 1])
            thresholds[0] = 0
            return thresholds

    @staticmethod
    def quant_act_scale(module: QuantHardTanh):
        bit_width = int(module.quant_act_bit_width().item())
        quant_act_scale = module.quant_act_scale().type(torch.FloatTensor).detach()
        if bit_width != 1:
            return quant_act_scale
        else:
            assert quant_act_scale.view(-1).shape[0] == 1, "Unsupported BIPOLAR per channel scale"
            assert quant_act_scale.flatten().item() == 1.0, "Unsupported BIPOLAR scale != 1"
            return quant_act_scale * 2

    def prepare_for_export(self, module: QuantHardTanh):
        self.symbolic_kwargs = {
            'qnt_type': self.quant_type(module),
            'thres': self.thresholds(module),
            'bias': self.quant_act_bias(module),
            'scale': self.quant_act_scale(module)}

    def symbolic_execution(self, inp: Tensor):
        ret = QuantHardTanhFn.apply(inp, *self.symbolic_kwargs.values())
        return ret


class FINNQuantIdentityHandler(FINNQuantHardTanhHandler):
    handled_layer = QuantIdentity



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from torch.autograd import Function

from brevitas.core.bit_width import BitWidthConst
from brevitas.core.function_wrapper.clamp import TensorClamp
from brevitas.core.quant import IntQuant
from brevitas.core.quant import TruncIntQuant
from brevitas.function import binary_sign
from brevitas.quant.solver.common import solve_float_to_int_impl_from_enum

DOMAIN_STRING = "onnx.brevitas"


class BrevitasBinaryQuantFn(Function):

    @staticmethod
    def symbolic(g, x, scale, zero_point, bit_width, narrow_range, signed, rounding_mode):
        ret = g.op(f'{DOMAIN_STRING}::BipolarQuant', x, scale)
        return ret

    @staticmethod
    def forward(ctx, x, scale, zero_point, bit_width, narrow_range, signed, rounding_mode):
        y = binary_sign(x) * scale
        return y


class BrevitasQuantFn(Function):

    @staticmethod
    def symbolic(g, x, scale, zero_point, bit_width, narrow_range, signed, rounding_mode):
        ret = g.op(
            f'{DOMAIN_STRING}::Quant',
            x,
            scale,
            zero_point,
            bit_width,
            rounding_mode_s=rounding_mode,
            signed_i=int(signed),
            narrow_i=int(narrow_range))
        return ret

    @staticmethod
    def forward(ctx, x, scale, zero_point, bit_width, narrow_range, signed, rounding_mode):
        float_to_int_impl = solve_float_to_int_impl_from_enum(rounding_mode)
        quant = IntQuant(
            float_to_int_impl=float_to_int_impl(),
            tensor_clamp_impl=TensorClamp(),
            narrow_range=narrow_range,
            signed=signed)
        y = quant(scale, zero_point, bit_width, x)
        return y


class BrevitasTruncFn(Function):

    @staticmethod
    def symbolic(g, x, scale, zero_point, input_bit_width, output_bit_width, rounding_mode):
        ret = g.op(
            f'{DOMAIN_STRING}::Trunc',
            x,
            scale,
            zero_point,
            input_bit_width,
            output_bit_width,
            rounding_mode_s=rounding_mode)
        return ret

    @staticmethod
    def forward(ctx, x, scale, zero_point, input_bit_width, output_bit_width, rounding_mode):
        float_to_int_impl = solve_float_to_int_impl_from_enum(rounding_mode)
        trunc = TruncIntQuant(
            float_to_int_impl=float_to_int_impl(),
            bit_width_impl=BitWidthConst(int(output_bit_width)))
        y_tuple = trunc(x, scale, zero_point, input_bit_width)
        return y_tuple[0]


class BrevitasQuantLSTMCellFn(Function):

    @staticmethod
    def symbolic(
            g,  # args and kwargs passed from _QuantLSTMLayer
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output,  # Symbolic kwargs passed from BrevitasQuantLSTMLayerHandler
            batch_first,
            reverse_input,
            cifg,  # Output quant
            output_scale,
            output_zero_point,
            output_bit_width,
            output_narrow_range,
            output_signed,
            output_rounding_mode,  # Cell state quant
            cell_state_scale,
            cell_state_zero_point,
            cell_state_bit_width,
            cell_state_narrow_range,
            cell_state_signed,
            cell_state_rounding_mode,  # Input gate accumulator quant
            input_acc_scale,
            input_acc_zero_point,
            input_acc_bit_width,
            input_acc_narrow_range,
            input_acc_signed,
            input_acc_rounding_mode,  # Forget gate accumulator quant
            forget_acc_scale,
            forget_acc_zero_point,
            forget_acc_bit_width,
            forget_acc_narrow_range,
            forget_acc_signed,
            forget_acc_rounding_mode,  # Cell gate accumulator quant
            cell_acc_scale,
            cell_acc_zero_point,
            cell_acc_bit_width,
            cell_acc_narrow_range,
            cell_acc_signed,
            cell_acc_rounding_mode,  # Output gate accumulator quant
            output_acc_scale,
            output_acc_zero_point,
            output_acc_bit_width,
            output_acc_narrow_range,
            output_acc_signed,
            output_acc_rounding_mode,  # Input gate sigmoid quant
            input_sigmoid_scale,
            input_sigmoid_zero_point,
            input_sigmoid_bit_width,
            input_sigmoid_narrow_range,
            input_sigmoid_signed,
            input_sigmoid_rounding_mode,  # Forget gate sigmoid quant
            forget_sigmoid_scale,
            forget_sigmoid_zero_point,
            forget_sigmoid_bit_width,
            forget_sigmoid_narrow_range,
            forget_sigmoid_signed,
            forget_sigmoid_rounding_mode,  # Cell gate tanh quant
            cell_tanh_scale,
            cell_tanh_zero_point,
            cell_tanh_bit_width,
            cell_tanh_narrow_range,
            cell_tanh_signed,
            cell_tanh_rounding_mode,  # Output gate sigmoid quant
            output_sigmoid_scale,
            output_sigmoid_zero_point,
            output_sigmoid_bit_width,
            output_sigmoid_narrow_range,
            output_sigmoid_signed,
            output_sigmoid_rounding_mode,  # Hidden state tanh quant
            hidden_state_tanh_scale,
            hidden_state_tanh_zero_point,
            hidden_state_tanh_bit_width,
            hidden_state_tanh_narrow_range,
            hidden_state_tanh_signed,
            hidden_state_tanh_rounding_mode):
        return g.op(
            f'{DOMAIN_STRING}::QuantLSTMCell',  # Tensors
            ## Input values
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output,  ## Output quant
            output_scale,
            output_zero_point,
            output_bit_width,  ## Cell state quant
            cell_state_scale,
            cell_state_zero_point,
            cell_state_bit_width,  ## Input gate accumulator quant
            input_acc_scale,
            input_acc_zero_point,
            input_acc_bit_width,  ## Forget gate accumulator quant
            forget_acc_scale,
            forget_acc_zero_point,
            forget_acc_bit_width,  ## Cell gate accumulator quant
            cell_acc_scale,
            cell_acc_zero_point,
            cell_acc_bit_width,  ## Output gate accumulator quant
            output_acc_scale,
            output_acc_zero_point,
            output_acc_bit_width,  ## Input gate sigmoid quant
            input_sigmoid_scale,
            input_sigmoid_zero_point,
            input_sigmoid_bit_width,  ## Forget gate sigmoid quant
            forget_sigmoid_scale,
            forget_sigmoid_zero_point,
            forget_sigmoid_bit_width,  ## Cell gate tanh quant
            cell_tanh_scale,
            cell_tanh_zero_point,
            cell_tanh_bit_width,  ## Output gate sigmoid quant
            output_sigmoid_scale,
            output_sigmoid_zero_point,
            output_sigmoid_bit_width,  ## Hidden state tanh quant
            hidden_state_tanh_scale,
            hidden_state_tanh_zero_point,
            hidden_state_tanh_bit_width,
            # Attributes
            batch_first_i=batch_first,
            reverse_input_i=reverse_input,
            cifg_i=cifg,
            output_narrow_i=output_narrow_range,
            output_signed_i=output_signed,
            output_rounding_mode_s=output_rounding_mode,
            cell_state_narrow_i=cell_state_narrow_range,
            cell_state_signed_i=cell_state_signed,
            cell_state_rounding_mode_s=cell_state_rounding_mode,
            input_acc_narrow_i=input_acc_narrow_range,
            input_acc_signed_i=input_acc_signed,
            input_acc_rounding_mode_s=input_acc_rounding_mode,
            forget_acc_narrow_i=forget_acc_narrow_range,
            forget_acc_signed_i=forget_acc_signed,
            forget_acc_rounding_mode_s=forget_acc_rounding_mode,
            cell_acc_narrow_i=cell_acc_narrow_range,
            cell_acc_signed_i=cell_acc_signed,
            cell_acc_rounding_mode_s=cell_acc_rounding_mode,
            output_acc_narrow_i=output_acc_narrow_range,
            output_acc_signed_i=output_acc_signed,
            output_acc_rounding_mode_s=output_acc_rounding_mode,
            input_sigmoid_narrow_i=input_sigmoid_narrow_range,
            input_sigmoid_signed_i=input_sigmoid_signed,
            input_sigmoid_rounding_mode_s=input_sigmoid_rounding_mode,
            forget_sigmoid_narrow_i=forget_sigmoid_narrow_range,
            forget_sigmoid_signed_i=forget_sigmoid_signed,
            forget_sigmoid_rounding_mode_s=forget_sigmoid_rounding_mode,
            cell_tanh_narrow_i=cell_tanh_narrow_range,
            cell_tanh_signed_i=cell_tanh_signed,
            cell_tanh_rounding_mode_s=cell_tanh_rounding_mode,
            output_sigmoid_narrow_range_i=output_sigmoid_narrow_range,
            output_sigmoid_signed_i=output_sigmoid_signed,
            output_sigmoid_rounding_mode_s=output_sigmoid_rounding_mode,
            hidden_state_tanh_narrow_i=hidden_state_tanh_narrow_range,
            hidden_state_tanh_signed_i=hidden_state_tanh_signed,
            hidden_state_tanh_rounding_mode_s=hidden_state_tanh_rounding_mode,
            # PyTorch requires to specify the number of outputs manually
            outputs=3)

    @staticmethod
    def forward(
            ctx,  # args and kwargs passed from _QuantLSTMLayer
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output,  # Symbolic kwargs passed from BrevitasQuantLSTMLayerHandler
            batch_first,
            reverse_input,
            cifg,  # Output quant
            output_scale,
            output_zero_point,
            output_bit_width,
            output_narrow_range,
            output_signed,
            output_rounding_mode,  # Cell state quant
            cell_state_scale,
            cell_state_zero_point,
            cell_state_bit_width,
            cell_state_narrow_range,
            cell_state_signed,
            cell_state_rounding_mode,  # Input gate accumulator quant
            input_acc_scale,
            input_acc_zero_point,
            input_acc_bit_width,
            input_acc_narrow_range,
            input_acc_signed,
            input_acc_rounding_mode,  # Forget gate accumulator quant
            forget_acc_scale,
            forget_acc_zero_point,
            forget_acc_bit_width,
            forget_acc_narrow_range,
            forget_acc_signed,
            forget_acc_rounding_mode,  # Cell gate accumulator quant
            cell_acc_scale,
            cell_acc_zero_point,
            cell_acc_bit_width,
            cell_acc_narrow_range,
            cell_acc_signed,
            cell_acc_rounding_mode,  # Output gate accumulator quant
            output_acc_scale,
            output_acc_zero_point,
            output_acc_bit_width,
            output_acc_narrow_range,
            output_acc_signed,
            output_acc_rounding_mode,  # Input gate sigmoid quant
            input_sigmoid_scale,
            input_sigmoid_zero_point,
            input_sigmoid_bit_width,
            input_sigmoid_narrow_range,
            input_sigmoid_signed,
            input_sigmoid_rounding_mode,  # Forget gate sigmoid quant
            forget_sigmoid_scale,
            forget_sigmoid_zero_point,
            forget_sigmoid_bit_width,
            forget_sigmoid_narrow_range,
            forget_sigmoid_signed,
            forget_sigmoid_rounding_mode,  # Cell gate tanh quant
            cell_tanh_scale,
            cell_tanh_zero_point,
            cell_tanh_bit_width,
            cell_tanh_narrow_range,
            cell_tanh_signed,
            cell_tanh_rounding_mode,  # Output gate sigmoid quant
            output_sigmoid_scale,
            output_sigmoid_zero_point,
            output_sigmoid_bit_width,
            output_sigmoid_narrow_range,
            output_sigmoid_signed,
            output_sigmoid_rounding_mode,  # Hidden state tanh quant
            hidden_state_tanh_scale,
            hidden_state_tanh_zero_point,
            hidden_state_tanh_bit_width,
            hidden_state_tanh_narrow_range,
            hidden_state_tanh_signed,
            hidden_state_tanh_rounding_mode):
        # Tp simplify things, here we are returning the outputs
        # as if they were already concatenated. Scale/zp/bw are avoided too.
        # This preserves output shapes but not values.
        # See _QuantLSTMCell for the actual implementation.
        quant_outputs = torch.zeros(
            quant_input.size(0),
            quant_input.size(1),
            quant_hidden_state.size(1),
            device=quant_hidden_state.device)
        return quant_outputs, quant_hidden_state, quant_cell_state
